# -*- coding: utf-8 -*-
"""
Created on Tue May 05 08:16:08 2015
@author: Ferriss

Diffusion in 1 and 3 dimensions with and without path-integration

### 1-dimensional diffusion ###
diffusionThinSlab and diffusion1D for profiles
With pynams profiles, use profile.plot_diffusion() and fitD()

### 3-dimensional diffusion without path integration: 3Dnpi ###
Simplest: diffusion3Dnpi(lengths, D's, time) to get a figure
    Step 1. Create parameters with params = params_setup3D
    Step 2. Pass parameters into diffusion3Dnpi_params(params) to get profiles
            Returns full 3D matrix v, sliceprofiles, then slice positions
    Step 3. Plot with styles.plot_3panels(slice positions, slice profiles)

### Whole-Block: 3-dimensional diffusion with path integration: 3Dwb ###
    Step 1. Create parameters with params = params_setup3D 
            Same as for non-path integrated 3D.
    Step 2. Pass parameters into diffusion3Dwb(params)
    Requires raypath information, unlike 3Dnpi

"""
from __future__ import print_function, division, absolute_import
import pynams.styles as styles
import numpy as np
import lmfit
import scipy
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1.parasite_axes import SubplotHost


#%% 1D Diffusion in a thin slab; Eq 4.18 in Crank, 1975
def diffusionThinSlab(log10D_m2s, thickness_microns, max_time_hours=2000, 
                      infinity=200, timesteps=300):
    """ 
    Eq 4.18 in Crank, 1975
    Takes log10 of the diffusivity D in m2/s, thickness in microns,
    and maximum time in hours, and returns time array in hours and 
    corresponding curve of C/C0, the concentration divided by the 
    initial concentration.
    """
    t_hours = np.linspace(0., max_time_hours, timesteps)
    t_seconds = t_hours * 3600.
    L_meters = thickness_microns / 2.E6
    cc = np.zeros_like(t_seconds)
    D_m2s = 10.**log10D_m2s

    for idx in range(len(t_seconds)):
        infsum = 0.
        for n in range(infinity):
            exponent_top = D_m2s * (((2.*n)+1.)**2.) * (np.pi**2.) * t_seconds[idx] 
            exponent_bot = 4. * (L_meters**2)
            exponent = np.exp(-1. * exponent_top / exponent_bot)
            addme = 8. * (1. / ((((2.*n)+1)**2.) * (np.pi**2.))) * exponent
            infsum = infsum + addme
        cc[idx] = infsum
    
    return t_hours, cc
    

#%% 1D diffusion profiles
def params_setup1D(microns, log10D_m2s, time_seconds, init=1., fin=0.,
                   vD=True, vinit=False, vfin=False, vTime=False):
    """
    Takes required info for diffusion in 1D - length, diffusivity, time,
    and whether or not to vary them - vD, vinit, vfin, vTime.
    
    Return appropriate lmfit params to pass into diffusion1D_params
    See https://lmfit.github.io/lmfit-py/parameters.html
    """
    params = lmfit.Parameters()
    params.add('microns', microns, False, None, None, None)
    params.add('log10D_m2s', log10D_m2s, vD, None, None, None)
    params.add('time_seconds', time_seconds, vTime, None, None, None)
    params.add('initial_unit_value', init, vinit, None, None, None)
    params.add('final_unit_value', fin, vfin, None, None, None)
    return params


def diffusion1D_params(params, 
                       data_x_microns=None, 
                       data_y_unit_areas=None, 
                       erf_or_sum='erf', 
                       centered=True, 
                       symmetric=True,
                       infinity=100, 
                       points=50):
    """
    Function set up to follow lmfit fitting requirements.
    See https://lmfit.github.io/lmfit-py/parameters.html
    
    Requires:
        lmfit parameters value dictionary 
        passing in key information as 'length_microns',
        'time_seconds', 'log10D_m2s', and 'initial_unit_value'. 
        See function params_setup1D

    If data are None (default), returns 1D unit diffusion profile 
    x_microns and y as vectors of length points (default 50). 

    Optional keywords:    
     - erf_or_sum: whether to use python's error functions (default) 
       or infinite sums
     - whether to center x data
     - whether the profile is symmetric or not. If not, change is to the right.
     - points sets how many points to calculate in profile. Default is 50.
     - what 'infinity' is if using infinite sum approximation
     
    If not including data, returns the x vector and model y values.
    With data, return the residual for use in fitting.
    """

    p = params.valuesdict()
    L_meters = p['microns'] / 1e6
    t = p['time_seconds']
    D = 10.**p['log10D_m2s']
    initial_value = p['initial_unit_value']
    final_value = p['final_unit_value']

    if initial_value > final_value:
        going_out = True
        solubility = initial_value
        minimum_value = final_value
    else:
        going_out = False        
        solubility = final_value
        minimum_value = initial_value
    
    a_meters = L_meters / 2.
    twoA = L_meters

    if t < 0:
        return           

    # Fitting to data or not? Default is not
    fitting = False
    if (data_x_microns is not None) and (data_y_unit_areas is not None):
        if len(data_x_microns) == len(data_y_unit_areas):
            fitting = True
        else:
            print('x and y data must be the same length')
            print('x', len(data_x_microns))
            print('y', len(data_y_unit_areas))
        
    # x is in meters and assumed centered around 0
    if fitting is True:            
        # Change x to meters and center it
        x = np.array(data_x_microns) / 1e6
        x = x - a_meters
    else:
        x = np.linspace(-a_meters, a_meters, points)
    
    # Make the infinite sum
    if erf_or_sum == 'infsum':
        xsum = np.zeros_like(x)
        for n in range(infinity):
           # positive number that converges to 1
            xadd1 = ((-1.)**n) / ((2.*n)+1.)        
            # time conponent
            xadd2 = np.exp(
                            (-D * (((2.*n)+1.)**2.) * (np.pi**2.) * t) / 
                            (twoA**2.) 
                            )                        
            # Where the position values come in to create the profile
            xadd3 = np.cos(
                            ((2.*n)+1.) * np.pi * x / twoA
                            )        
            xadd = xadd1 * xadd2 * xadd3
            xsum = xsum + xadd
            
        model = xsum * 4. / np.pi
    
    # or make the error function
    elif erf_or_sum == 'erf':
        if symmetric is False:
            a_meters = a_meters*2
            x = x*2
            
        sqrtDt = (D*t)**0.5
        model = ((scipy.special.erf((a_meters+x)/(2*sqrtDt))) + 
                   (scipy.special.erf((a_meters-x)/(2*sqrtDt))) - 1) 
        
    else:
        print ('erf_or_sum must be set to either "erf" for python built-in ' +
               'error function approximation (defaul) or "sum" for infinite ' +
               'sum approximation with infinity=whatever, defaulting to ' + 
               str(infinity))
        return False

    if going_out is False:
        model = np.ones_like(model) - model

    concentration_range = solubility - minimum_value
    model = (model * concentration_range) + minimum_value

    x_microns = x * 1e6
    
    # If not including data, just return the model values
    # With data, return the residual for use in fitting.
    if fitting is False:
        return x_microns, model
    return model- data_y_unit_areas


def diffusion1D(length_microns, log10D_m2s, time_seconds, init=1., fin=0.,
                erf_or_sum='erf', show_plot=True, 
                style=styles.style_blue, infinity=100, points=100, 
                centered=True, axes=None, symmetric=True,
                maximum_value=1.):
    """
    Simplest implementation for 1D diffusion.
    
    Takes required inputs length, diffusivity, and time 
    and plots diffusion curve on new or specified figure. 
    Optional inputs are unit initial value and final values. 
    Defaults assume diffusion out, so init=1. and fin=0. 
    Reverse these for diffusion in.
    
    Change scale of y-values with maximum_value keyword.
    
    Returns figure, axis, x vector in microns, and model y data.
    """    
    if symmetric is True:
        params = params_setup1D(length_microns, log10D_m2s, time_seconds,
                                init=init, fin=fin)
        x_diffusion, y_diffusion = diffusion1D_params(params, points=points)
        if centered is False:
            a_length = (max(x_diffusion) - min(x_diffusion)) / 2
            x_diffusion = x_diffusion + a_length
    else:
        # multiply length by two
        params = params_setup1D(length_microns*2, log10D_m2s, time_seconds,
                                init=init, fin=fin)
        x_diffusion, y_diffusion = diffusion1D_params(params, points=points)        

        # divide elongated profile in half
        x_diffusion = x_diffusion[int(points/2):]
        y_diffusion = y_diffusion[int(points/2):]
        if centered is True:
            a_length = (max(x_diffusion) - min(x_diffusion)) / 2
            x_diffusion = x_diffusion - a_length 

    if show_plot is True:
        if axes is None:
            fig = plt.figure()          
            ax  = SubplotHost(fig, 1,1,1)
            ax.grid()
            ax.set_ylim(0, maximum_value)
            ax.set_xlabel('position ($\mu$m)')
            ax.set_xlim(min(x_diffusion), max(x_diffusion))
            ax.plot(x_diffusion, y_diffusion*maximum_value, **style)
            ax.set_ylabel('Unit concentration or final/initial')
            fig.add_subplot(ax)
        else:
            axes.plot(x_diffusion, y_diffusion*maximum_value, **style)
            fig = None
            ax = None            
    else:
        fig = None
        ax = None
    
    return fig, ax, x_diffusion, y_diffusion


#%% 3-dimensional diffusion parameter setup
def length_checker(microns3):
    """
    Checks lengths passed in, and returns possibly modified lengths
    """
    if isinstance(microns3, int):
        microns3 = [float(microns3)]*3
    elif isinstance(microns3, float):
        microns3 = [microns3]*3
    elif isinstance(microns3, list):
        if len(microns3) == 1:
            microns3 = microns3*3
        elif len(microns3) == 2:
            new_length = np.mean(microns3)
            print('Warning: assuming 3rd length=', new_length)
            microns3 = microns3.append(new_length)
    else:
        print('only int, float, or list allowed for lengths_microns')
        return False
    return microns3


def D_checker(log10D3):
    """
    Checks diffusivities passed in and returns possibly modified diffusivities.
    """
    if isinstance(log10D3, int):
        log10D3 = [float(log10D3)]*3
    elif isinstance(log10D3, float):
        log10D3 = [log10D3]*3
    elif isinstance(log10D3, list):
        if len(log10D3) == 1:
            log10D3 = log10D3*3
        elif len(log10D3) == 2:
            new_D = np.mean(log10D3)
            print('Warning: assuming 3rd D = log10', new_D)
            log10D3 = log10D3.append(new_D)
    else:
        print('only int, float, or list allowed for log10Ds_m2s')
        return False
    return log10D3


def params_setup3D(microns3, log10D3, time_seconds, 
                   initial=1., final=0., isotropic=False, slowb=False,
                   vD=[True, True, True], vinit=False, vfin=False):
    """
    Requires:
        microns3: a list of 3 lengths in microns
        log10D3: a list of 3 diffusivities in m2/s and converted to log10
        time_seconds: the time in seconds
        
    Optional input:
        initial and final unit values
        whether to force diffusive anisotropy (default: isotropic=False)
        slowb: whether to force [010] to be an order of magnitude slower
        vD: a list of whether to vary diffusivities and which ones
        vinit: whether to vary the initial concentration
        vfin: whether to vary the final concentration
    
    Returns:
        an lmfit parameters object for use in 3D fitting functions
        See https://lmfit.github.io/lmfit-py/parameters.html

    """
    params = lmfit.Parameters()
    params.add('microns3', microns3, False, None, None, None)
    params.add('log10Dx', float(log10D3[0]), vD[0], None, None, None)
    params.add('time_seconds', float(time_seconds), False, None, None, None)
    params.add('initial_unit_value', float(initial), vinit, None, None, None)
    params.add('final_unit_value', float(final), vfin, None, None, None)

    if isotropic is True:
        params.add('log10Dy', float(log10D3[1]), vD[1], None, None, 'log10Dx')
        params.add('log10Dz', float(log10D3[2]), vD[2], None, None, 'log10Dx')
    elif slowb is True:
        params.add('log10Dy', float(log10D3[1]), vD[1], None, None,\
                   'log10Dx - 1.')
        params.add('log10Dz', float(log10D3[2]), vD[2], None, None, 'log10Dx')
    else:            
        params.add('log10Dy', float(log10D3[1]), vD[1], None, None, None)
        params.add('log10Dz', float(log10D3[2]), vD[2], None, None, None)

    return params


def diffusion3Dnpi_params(params, 
                          data_x_microns=None, 
                          data_y_unit_areas=None, 
                          erf_or_sum='erf', 
                          centered=True, 
                          infinity=100, 
                          points=50):
    """ 
    Diffusion in 3 dimensions in a rectangular parallelipiped.
    Requires:
        params - lmfit parameters. Setup parameters with params_setup3D.
        For more on parameters, see 
        https://lmfit.github.io/lmfit-py/parameters.html 
    
    Option inpur aimilar to diffusion1D_params.
    
    Returns:
        v: complete 3D concentration matrix
        y: slice profiles
        x: positions of slice profiles
    
    ### NOT SET UP FOR FITTING YET ###
    """   
    
    fitting = False
    
    if (data_x_microns is not None) and (data_y_unit_areas is not None):
        x_data = np.array(data_x_microns)
        y_data = np.array(data_y_unit_areas)
        
        if np.shape(x_data) == np.shape(y_data):
            fitting = True
            print('fitting to data')
        else:
            print('x and y data must be the same shape')
            print('x', np.shape(x_data))
            print('y', np.shape(y_data))

    p = params.valuesdict()
    L3_microns = np.array(p['microns3'])
    t = p['time_seconds']
    init = p['initial_unit_value']
    vary_init = [params['initial_unit_value'].vary]
    fin = p['final_unit_value']
    vary_fin = [params['final_unit_value'].vary]
    log10D3 = [p['log10Dx'], p['log10Dy'], p['log10Dz']]
    vary_D = [params['log10Dx'].vary, 
              params['log10Dy'].vary, 
              params['log10Dz'].vary]

    # going in or out?
    if init < fin: 
        going_out = False
        init, fin = fin, init
    else:
        going_out = True
       
    # First create 3 1D profiles, 1 in each direction
    xprofiles = []    
    yprofiles = []
    kwdict = {'points' : points}
    
    for microns, D, vD in zip(L3_microns, log10D3, vary_D):
        p1D = lmfit.Parameters()
        p1D.add('microns', float(microns), False)
        p1D.add('time_seconds', t, params['time_seconds'].vary)
        p1D.add('log10D_m2s', D, vD)
        p1D.add('initial_unit_value', 1., vary_init)
        p1D.add('final_unit_value', 0., vary_fin)
        
        try:
            x, y = diffusion1D_params(p1D, **kwdict)
        except TypeError:
            print('Problem with x values after diffusion1D')
            print(diffusion1D_params(p1D, **kwdict))
            return

        xprofiles.append(x)
        yprofiles.append(y)
                                      
    # Then multiply them together to get a 3D matrix
    # I should figure out how to do this without the for-loops
    v = np.ones((points, points, points))
    for d in range(0, points):
        for e in range(0, points):
            for f in range(0, points):
                v[d][e][f] = yprofiles[0][d]*yprofiles[1][e]*yprofiles[2][f]

    if going_out is False:
        v = np.ones((points, points, points)) - v

    scale = np.abs(fin-init)
    minimum_value = min([fin, init])
    v = (v * scale) + minimum_value

    mid = int(points/2.)    
    aslice = v[:, mid][:, mid]
    bslice = v[mid][:, mid]
    cslice = v[mid][mid]
    sliceprofiles = [aslice, bslice, cslice]

    slice_positions_microns = []
    for k in range(3):
        a = L3_microns[k] / 2.
        if centered is False:            
            x = np.linspace(0, a*2., points)
        else:
            x = np.linspace(-a, a, points)
        slice_positions_microns.append(x)
          
    # Returning full matrix and 
    # slice profiles in one long list for use in fitting
    sliceprofiles = [aslice, bslice, cslice]
    
    if fitting is False:
        return v, sliceprofiles, slice_positions_microns
    else:
        ### Still need to set up residuals! ###
        residuals = np.zeros_like(sliceprofiles)
        return residuals


def diffusion3Dnpi(lengths_microns, log10Ds_m2s, time_seconds, points=50,
                   initial=1, final=0., ytop=None, show_plot=True, 
                   centered=True, styles3=[None]*3, axes=None, 
                   show_line_at_initial=True):
        """
        Required input:
        list of 3 lengths, list of 3 diffusivities, and time 
        
        Optional input:
        initial unit concentration (default initial=1), 
        final unit concentration (default final=0), 
        whether to plot output (default show_plot=True), 
        maximum y limit on plot (ytop),
        and number of points to use during calculation (default points=50)
        
        If show_plot=True (default), plots results and returns:
        1. f, figure of plot of 3D non-path-averaged diffusion profiles.
        2. ax, 3 axes of figure
        3. v, 3D matrix of diffusion
        4. x, list of 3 sets of x values plotted
        5. y, list of 3 sets of y values plotted
        
        If show_plot=False, returns only v, x, y
        """       
        lengths_microns = length_checker(lengths_microns)
        log10Ds_m2s = D_checker(log10Ds_m2s)
        params = params_setup3D(lengths_microns, log10Ds_m2s, time_seconds,
                                initial=initial, final=final)
        
        v, y, x = diffusion3Dnpi_params(params, points=points, 
                                        centered=centered)

        if show_plot is True:
            try:
                if axes is None:
                    f, ax = styles.plot_3panels(x, y, figaxis3=axes, 
                                        ytop=ytop, centered=centered,
                                        styles3=styles3, init=initial,
                                        lengths=lengths_microns,
                                        show_line_at_1=show_line_at_initial)
                    return f, ax, v, x, y
                else:
                    styles.plot_3panels(x, y, ytop=ytop, centered=centered, 
                                        styles3=styles3, figaxis3=axes,
                                        lengths=lengths_microns,
                                        show_line_at_1=show_line_at_initial,
                                        init=initial)
                    return v, x, y
            except(TypeError):
                print
                print('TypeError: problem in plot_3panels()')
                return v, x, y
                
        else:
            return v, x, y
            
        
#%% 3D whole-block: 3-dimensional diffusion with path integration
def diffusion3Dwb_params(params, data_x_microns=None, data_y_unit_areas=None, 
                          raypaths=None, erf_or_sum='erf', show_plot=True, 
                          fig_ax=None, style=None, need_to_center_x_data=True,
                          infinity=100, points=50, show_1Dplots=False):
    """ 
    Diffusion in 3 dimensions with path integration.
    Requires setup with params_setup3Dwb
    """
    if raypaths is None:
        print('raypaths must be in the form of a list of three abc directions')
        return

    v, y, x = diffusion3Dnpi_params(params, points=int(points), 
                                    centered=False)
    
    # Fitting to data or not? Default is not
    # Add appropriate x and y data to fit
    fitting = False
    if (data_x_microns is not None) and (data_y_unit_areas is not None):
        x_array = np.array(data_x_microns)
        y_array = np.array(data_y_unit_areas)
        if np.shape(x_array) == np.shape(y_array):
            print('fitting to data')
            fitting = True
        else:
            print('x and y data must be the same shape')
            print('x', np.shape(x_array))
            print('y', np.shape(y_array))
            
    # Whole-block measurements can be obtained through any of the three 
    # planes of the whole-block, so profiles can come from one of two ray path
    # directions. These are the planes.
    raypathA = v.mean(axis=0)
    raypathB = v.mean(axis=1)
    raypathC = v.mean(axis=2)
    
    # Specify whole-block profiles in model
    mid = int(points/2)
    if raypaths[0] == 'b':
        wbA = raypathB[:, mid]
    elif raypaths[0] == 'c':
        wbA = raypathC[:, mid]       
    else:
        print('raypaths[0] for profile || a must be "b" or "c"')
        return
        
    if raypaths[1] == 'a':
        wbB = raypathA[:, mid]
    elif raypaths[1] == 'c':
        wbB = raypathC[mid]       
    else:
        print('raypaths[1] for profile || b must be "a" or "c"')
        return

    if raypaths[2] == 'a':
        wbC = raypathA[mid]
    elif raypaths[2] == 'b':
        wbC = raypathB[mid]       
    else:
        print('raypaths[2] for profile || c must be "a" or "b"')
        return

    p = params.valuesdict()
    L3 = p['microns3']
    
    wb_profiles = [wbA, wbB, wbC]
    wb_positions = []
    for length in L3:
        x_microns = np.linspace(0., float(length), points)
        wb_positions.append(x_microns)
        
    if show_plot is True:
        if style is None:
            style = [None, None, None]
            for k in range(3):
                style[k] = styles.style_lightgreen

        if fig_ax is None:
            f, fig_ax = styles.plot_3panels(wb_positions, wb_profiles, 
                                            L3, style)
        else:
            styles.plot_3panels(wb_positions, wb_profiles, L3, style, 
                         figaxis3=fig_ax)                         

    if fitting is False:        
        return wb_positions, wb_profiles
    
    if fitting is True:
        # Return residuals 
        y_model = []
        y_data = []
        residuals = []
        for k in range(3):
            for pos in range(len(x_array[k])):
                # wb_positions are centered, data are not
                microns = x_array[k][pos]
                
                # Find the index of the full model whole-block value 
                # closest to the data position
                idx = (np.abs(wb_positions[k]-microns).argmin())
                
                model = wb_profiles[k][idx]
                data = y_array[k][pos]
                res = model - data
                
                y_model.append(model)
                y_data.append(data)
                residuals.append(res)                
        return np.array(residuals)


def diffusion3Dwb(lengths_microns, log10Ds_m2s, time_seconds, raypaths,
                   initial=1., final=0., ytop=1.2, points=50, show_plot=True,
                   axes=None, isotropic=False, centered=True,
                   show_line_at_initial=True, styles3=[None]*3):
        """
        Takes list of 3 lengths, list of 3 diffusivities, and time.
        Returns plot of 3D path-averaged (whole-block) diffusion profiles.
        
        Requirements and keywords similar to diffusion.3Dnpi except a list
        of raypaths (thickness directions as 'a', 'b', or 'c')
        """
        lengths_microns = length_checker(lengths_microns)
        log10Ds_m2s = D_checker(log10Ds_m2s)
        params = params_setup3D(lengths_microns, log10Ds_m2s, time_seconds,
                                initial=initial, final=final)
        
        x, y = diffusion3Dwb_params(params, raypaths=raypaths, show_plot=False,
                                    points=int(points))

        if centered is True:
            for idx in range(3):
                x[idx] = x[idx] - (lengths_microns[idx] / 2.)
                        
        if show_plot is True:
            if axes is None:
                f, ax = styles.plot_3panels(x, y, figaxis3=axes, 
                                        ytop=ytop, centered=centered,
                                        styles3=styles3, init=initial,
                                        lengths=lengths_microns,
                                        show_line_at_1=show_line_at_initial)

                return f, ax, x, y
            else: 
                styles.plot_3panels(x, y, figaxis3=axes, 
                                    ytop=ytop, centered=centered,
                                    styles3=styles3, init=initial,
                                    lengths=lengths_microns,
                                    show_line_at_1=show_line_at_initial)

        return x, y


# -*- coding: utf-8 -*-
"""
Created on Mon Dec 19 14:34:00 2016

Define attributes and functions related to physical samples

@author: Ferriss
"""
from __future__ import print_function, division, absolute_import
import numpy as np

class Sample():
    """
    Stores basic sample geometry. Most sample info should be stored with 
    the IGSN, the international geosample number, available at geosamples.org.
    You can include the IGSN and/or a description of the sample when 
    setting up the sample as well, e.g.,
    my_sample = Sample(IGSN=IEFERJAI4, description='San Carlos olivine')

    For a thin slab where you're not worried about the direction, 
    input thickness under thickness_thinslab_microns, e.g., 
    my_thin_slab = Sample(thickness_thinslab_microns=200.)

    or with multiple measurements, use a list of numbers, e.g.,
    my_thin_slab = Sample(thickness_thinslab_microns=[212., 206., 230.])
    
    For a sample block, measure or guess the orientation and input one or 
    more measurements under length_a/b/c_microns, e.g.,
    my_block = Sample(length_a_microns=1000,
                      length_b_microns=[2012, 2020],
                      length_c_microns=[2410, 2400])

    All thicknesses input should be in microns.
    
    The Sample will automatically generate an attribute thickness_microns that
    is a list of the average thickness || a, b, c, and unoriented.
    e.g., print my_block.thickness_microns produces
    [1000.0, 2016.0, 2405.0, None]
    
    Fe (total), Fe2+, Fe3_, Mg, Al, and Ti concentratrations can be added 
    in as atoms per formula unit (apfu), and initial water content can be
    added in ppm H2O.
    Sample(Fe2=0.17, initial_water_ppmH2O=5)
    
    You can have it determine the Mg# from the Fe and Mg
    by these using Sample.get_MgNum()
    """
    def __init__(self, IGSN=None, 
                 description=None,
                 thickness_thinslab_microns=[],
                 length_a_microns = [],
                 length_b_microns = [], 
                 length_c_microns = [],                  
                 Fe=0, Fe2=0, Fe3=0., Mg=0, Al=0, Ti=0,
                 initial_water_ppmH2O=None):
        self.IGSN = IGSN
        self.description = description
        self.initial_water_ppmH2O = initial_water_ppmH2O
        self.length_a_microns = length_a_microns
        self.length_b_microns = length_b_microns
        self.length_c_microns = length_c_microns
        self.thicknesses_microns = [length_a_microns, 
                                    length_b_microns,
                                    length_c_microns]
        self.Fe2 = Fe2
        self.Fe3 = Fe3
        self.Mg = Mg
        self.Al = Al
        self.Ti = Ti
        if (Fe2 is not None) and (Fe3 is not None) and (Fe is None):
            self.Fe = Fe2 + Fe3
        else:
            self.Fe = Fe


        def floatify(lengthy):
            """
            Takes a length or list of lengths and makes it into a single 
            floating point number for condensing down lists of thicknesses.
            """
            if lengthy is None:
                floater = None
            elif isinstance(lengthy, list) or isinstance(lengthy, np.ndarray):
                if len(lengthy) > 0:
                    floater = np.mean(lengthy)
                else:
                    floater = None
            else:
                try:
                    floater = float(lengthy)
                except TypeError:
                    floater = lengthy
                    print('length should be integer, float, or list')
            return floater

        twoA = floatify(length_a_microns)
        twoB = floatify(length_b_microns)
        twoC = floatify(length_c_microns)
        thinness = floatify(thickness_thinslab_microns)
        self.thickness_microns = [twoA, twoB, twoC, thinness]
        self.lengths_microns = [twoA, twoB, twoC]

        
    def get_MgNumber(self):
        """ 
        Takes the Fe (or Fe2 + Fe3 if Fe=0) and Mg of a sample assumed to be 
        in atoms per formula unit (apfu) and calculates the Mg#
        """
        if self.Fe == 0:
            self.Fe = self.Fe2 + self.Fe3
        if self.Fe == 0 and self.Mg == 0:
            print("You don't seem to have any Fe or Mg")
            return
        if self.Fe == 0:
            print("You don't seem to have any Fe")
        if self.Mg == 0:
            print("You don't seem to have any Mg")            
        try:
            MgNum = 100. * self.Mg / (self.Fe + self.Mg)
            print('\nThe Mg#:')
            print(MgNum)
            self.MgNum = MgNum
        except TypeError:
            print('\nCheck your Mg and Fe values')


# -*- coding: utf-8 -*-
"""
Created on Tue May 05 08:16:08 2015
@author: Ferriss

class Diffusivities() groups together temperatures and diffusivities
for use in plotting directly onto Arrhenius diagrams and solving 
for activation energies and pre-exponential components

Also provides basic plotting and functions for creating Arrhenius diagrams and 
handling groups of diffusivities that get plotted on them, including 
determining activation energies and pre-exponential terms and calculating
the expected diffusivity at a given temperature.

"""
from __future__ import print_function, division, absolute_import
from pynams import styles
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.lines as mlines
from mpl_toolkits.axes_grid1.parasite_axes import SubplotHost
from uncertainties import ufloat
import sys

GAS_CONSTANT = 0.00831 # kJ/mol K


def solve_Ea_D0(log10D_list, celsius_list, printout=True):
    """
    Takes lists of diffusivities as log10 in m2/s and associated 
    temperatures in celsius. Returns activation energy Ea in kJ/mol K and D0 
    in m2/s. The errors on the individual diffusivities are not included.
    """
    T = np.array(celsius_list) + 273.15
    x = 1.E4 / T
    y = np.array(log10D_list)

    try:
        # If I don't add in a very low weighted extra number, the covariance
        # matrix, and hence the error, comes out as infinity. The actual
        # fitting results don't change. The very low number in this case I'm 
        # getting from sys.float_info.
        x_extra = np.concatenate((x, x[-1:]), axis=0)
        y_extra = np.concatenate((y, y[-1:]), axis=0)    
        weights = list(np.ones(len(x))) + [sys.float_info.epsilon]
        try:
            fit_extra, cov_extra = np.polyfit(x_extra, 
                                              y_extra, 1, 
                                              w=weights, 
                                              cov=True)
        except np.linalg.linalg.LinAlgError as err:
            return 0, 0                            
            
        if cov_extra[0][0] > 0:
            Ea_error = cov_extra[0][0]
        else:
            Ea_error = 0.
        if cov_extra[1][1] > 0:
            D0_error = cov_extra[1][1]
        else:
            D0_error = 0.
            
        Ea = -ufloat(fit_extra[0], Ea_error) * 2.303 * GAS_CONSTANT * 1.E4
        D0 = 10.**ufloat(fit_extra[1], D0_error)
        
    except ValueError:
        fit_extra = np.polyfit(x_extra, y_extra, 1)
        Ea = fit_extra[0] * 2.303*GAS_CONSTANT*1.E4
        D0 = 10.**fit_extra[1]
    except IndexError:
        # probably no points or unequal number of points
        Ea = 0
        D0 = 0
    except TypeError:
        Ea = 0
        D0 = 0
    
    if printout is True:
        print(Ea, D0)
        
    return Ea, D0


def whatIsD(Ea, D0, celsius, printout=True):
    """
    Takes activation energy in kJ/mol, D0 in m2/s and 
    temperature in celsius. Returns log10 diffusivity in m2/s
    """
    T = celsius + 273.15
    try:
        D0 = D0.n
    except AttributeError:
        pass
    
    try:
        Ea = Ea.n
    except AttributeError:
        pass
    
    D = D0 * np.exp(-Ea / (GAS_CONSTANT * T))
    D = np.log10(D)
        
    if printout is True:
        print('log10 D at ', celsius, 'C: ', '{:.1f}'.format(D), ' in m2/s')
    return D


class Diffusivities():
    def __init__(self, 
                description=None, 
                celsius = [[], [], [], []],
                log10D = [[], [], [], []],
                activation_energy_kJmol = [],
                D0_m2s = [],
                sample=None,  
                ):
        """
        Handy groupings of experimentally determined diffusivities.
        
        Data is grouped as lists in order or orientation:
        parallel a, parallel b, parallel c, and unoriented.
        """
        self.description = description
        self.celsius = celsius
        self.log10D = log10D
        self.sample = sample
        self.activation_energy_kJmol = activation_energy_kJmol
        self.D0_m2s = D0_m2s


    def fill_in_data(self, df, mech, percentpv, paper=None):
        """
        See usage example in literaturevalues.py in this folder.
        
        Input: grouped pandas dataframe, 
               mechanism name (bulk, [Mg], etc.)
               the paper name used to make the pandas groupby
               default assumes paper = self.description
        
        Fills in temperature and diffusivities from a spreadsheet.
        """
        self.log10D = [[], [], [], []]
        self.celsius = [[], [], [], []]
        if paper is None:
            paper = self.description
        for idx, orient in enumerate(['a', 'b', 'c', 'u']):
            group = (paper, mech, float(percentpv), orient)
            try:
                self.log10D[idx] = list(df.get_group(group)['log10D'])                
            except KeyError:
                self.log10D[idx] = []
            
            try:
                self.celsius[idx] = list(df.get_group(group)['celsius'])
            except KeyError:
                self.celsius[idx] = []
       
        
    def solve_Ea_D0(self, printout=True):
        """
        Returns and saves as attributes the
        activation energy in kJ/mol and D0 (log10) in m2/s 
        """ 
        self.activation_energy_kJmol = []
        self.D0_m2s = []
        for idx in range(4):
            logD = self.log10D[idx]
            celsius = self.celsius[idx]
            Ea, D0 = solve_Ea_D0(logD, celsius, printout=printout)
            self.activation_energy_kJmol.append(Ea)
            self.D0_m2s.append(D0)


    def whatIsD(self, celsius, orient='ALL', printout=True):
        """ 
        Takes temperature in celsius. 
        
        Returns log10 diffusivity in m2/s.
        """ 
        D = [] 
        for idx, direction in enumerate(['a', 'b', 'c', 'not oriented']):
            try:
                Ea = self.activation_energy_kJmol[idx]
                D0 = self.D0_m2s[idx]
                
            except IndexError:
                print('\nThese should have 4 items')
                print('Remember 0 placeholder for no measurements')
                print('activation energy')
                print(self.activation_energy_kJmol)
                print('D0')
                print(self.D0_m2s)
                return None
            
            try:
                if (D0 != 0) and (Ea != 0):
                    xD = whatIsD(Ea, D0, celsius, printout=False)
                else: 
                    xD = 'D0 or Ea = 0'
            except TypeError:
                xD = 'Type Error'
                
            if (printout is True) and (orient == 'ALL'):
                print(xD, '||', direction)
            D.append(xD)
        
        orient_list = ['a', 'b', 'c', 'u', 'not oriented', 'ALL']
        if orient in orient_list:
            idx = orient_list.index(orient)
            if idx == 4:
                idx = 3
            if (printout is True) and (orient != 'ALL'):
                print(D[idx], '||', orient)
        else:
            print('Accepted orient values:')
            print('ALL default', orient_list)
        return D
        

    def plotD(self, axes, orient='ALL', plotdata=True,
              offset_celsius=0, plotline=True, extrapolate_line=False,
              legend_add=False, 
              legend_handle=None, style=styles.style_points.copy(), 
              error_color=None, lower_legend_by=-2.0,
              style_line=styles.style_1, label=None):
        """
        Takes axes handle for Arrhenius diagram created by 
        Arrhenius_outline() below and plots data (plotdata=True) and 
        best-fit line (plotline=True) for specified orientations in
        orient (default='ALL'). 
        
        If extrapolate_line=True, the line will be extended all the way 
        to the edges of the plot. 
        
        You can also add it to the legend (default) or not (legend_add=False).
        If you are adding to the legend, you *must* also pass in the 
        legend_handle, which was returned by Arrhenius_outline()
        
        """
        if orient == 'ALL':
            orient_list = list(range(4))
        else:
            iorient = styles.get_iorient(orient)    
            orient_list = [iorient]
                  
        for iorient in orient_list:
            celsius = self.celsius[iorient]
            logD = self.log10D[iorient]

            if orient == 'ALL':
                label = None
                
            if label is None:
                if iorient == 0:
                    label = '|| [100]'
                elif iorient == 1:
                    label = '|| [010]'
                elif iorient == 2:
                    label = '|| [001]'
                elif iorient == 3:
                    label = 'not oriented'                   
  
            if legend_add is True and legend_handle is None:
                print(self.description)
                print('Need legend_handle for legend')
                return
               
            if (len(celsius) == 0) and (self.celsius_all is not None):
                celsius = self.celsius_all

            if (len(celsius) == 0):
                continue
    
            if logD is None:
                continue
    
            if len(celsius) != len(logD):
                print('\n', self.description)
                print('temp and logD not the same length')
                continue
    
            # change temperature scale                   
            x = []
            for k in range(len(celsius)):
                x.append(1.0e4 / (celsius[k] + offset_celsius + 273.15))

            if plotline is True:    
                Tmin = min(x)
                Tmax = max(x)
                T = [Tmin, Tmax]
                p = np.polyfit(x, logD, 1)
                
                if extrapolate_line is True:
                    extrap = axes.get_xlim()
                    axes.plot(extrap,np.polyval(p, extrap), **style_line)
                else:
                    axes.plot(T,np.polyval(p, T), **style_line)
                   
            if label is not None:
                style['label'] = label
            if plotdata is True:
                axes.plot(x, logD, **style)
            
        if legend_add is True:
            self.add_to_legend(axes, legend_handle, style=style,
                               style_line=style_line, plotline=plotline,
                               lower_legend_by=lower_legend_by)

            
    def add_to_legend(self, axes, legend_handle, lower_legend_by=-2.0,
                      orient=None, plotline=False,
                      ncol=2, style=styles.style_points.copy(), 
                      style_line=styles.style_1, label=None):
        """
        Take a figure axis and its list of legend handles 
        and adds information to it
        """
        if label is None:
            if self.description is None:
                print('Need label or self.description to make legend')
                return
            else:
               descript = self.description
        else:
            descript = label
        style['label'] = descript

        add_marker = mlines.Line2D([], [], **style) 
        
        legend_handle.append(add_marker)
        
        low = axes.get_xlim()[0]
        high = axes.get_xlim()[1]
        bottom = axes.get_ylim()[0]
        main_legend = plt.legend(handles=legend_handle, 
                                 numpoints=1, ncol=ncol, 
                                 bbox_to_anchor=(low, bottom, high-low, 
                                                 lower_legend_by),
                                 bbox_transform=axes.transData, 
                                 mode='expand')
        plt.gca().add_artist(main_legend)
        return legend_handle

generic = Diffusivities()
generic.basestyle = {'marker' : 's', 'color' : 'black', 'alpha' : 0.5,
                     'markersize' : 8, 'linestyle': 'none'}


def Arrhenius_outline(xlow=6., xhigh=11., ybottom=-18., ytop=-8.,
                      celsius_labels = np.arange(0, 2000, 100),
                      shrink_axes_to_fit_legend_by = 0.3, make_legend=False,
                      lower_legend_by=-2., ncol=2):
    """
    Make Arrhenius diagram outline. 
    
    Returns figure, axis, legend handle.
    
    low, high, top, and bottom set the x and y axis limits. 

    celsius_labels sets where to make the temperature tick marks.
    
    If you have issues with the legend position or overlap with main diagram,
    play with the numbers for shrink_legend_by and lower_legend_by
    
    ncol sets the number of columns in the legend.
    """
    fig = plt.figure()
    ax = SubplotHost(fig, 1,1,1)
    ax_celsius = ax.twin()
    parasite_tick_locations = 1e4/(celsius_labels + 273.15)
    ax_celsius.set_xticks(parasite_tick_locations)
    ax_celsius.set_xticklabels(celsius_labels)
    fig.add_subplot(ax)
    ax.axis["bottom"].set_label("10$^4$/Temperature (K$^{-1}$)")
    ax.axis["left"].set_label("log$_{10}$diffusivity (m$^{2}$/s)")
    ax_celsius.axis["top"].set_label("Temperature ($\degree$C)")
    ax_celsius.axis["top"].label.set_visible(True)
    ax_celsius.axis["right"].major_ticklabels.set_visible(False)
    ax.set_xlim(xlow, xhigh)
    ax.set_ylim(ybottom, ytop)
    ax.grid()
    
    # main legend below
    if make_legend is True:
        legend_handles_main = []
        box = ax.get_position()
        ax.set_position([box.x0, box.y0 + box.height*shrink_axes_to_fit_legend_by, 
                         box.width, box.height*(1.0-shrink_axes_to_fit_legend_by)])
        main_legend = plt.legend(handles=legend_handles_main, numpoints=1, 
                                 ncol=ncol, 
                                 bbox_to_anchor=(xlow, ybottom, xhigh-xlow, 
                                                 lower_legend_by),
                                 bbox_transform=ax.transData, mode='expand')
        plt.gca().add_artist(main_legend)
    else:
        legend_handles_main = None
    return fig, ax, legend_handles_main


def Arrhenius_add_line(fig_ax, Ea, D0, xlow=6.0, xhigh=10.0, 
                       style={'color' : 'k', 'linestyle' : '-'}):
    """
    Takes figure axis from above, Ea activation energy in kJ/mol, D0 in m2/s
    Plots Arrhenius line from 1E4/T = xlow to xhigh
    """
    T = 1E4 / np.linspace(xlow, xhigh) 
    log10D = np.log10(D0) - (Ea/(2.303 * GAS_CONSTANT * T))
    fig_ax.plot(1E4 / T, log10D, **style)




"""
Code for grouping 3 orthogonal Profiles into a single Block object.
"""
from __future__ import print_function, division, absolute_import
import pynams.styles as styles
from pynams.diffusion import models
from pynams import Spectrum
import numpy as np
import matplotlib.pyplot as plt
import lmfit

class Block():
    def __init__(self,profiles=[], folder='', name='', get_peakfit=False, 
                 time_seconds=None, sample=None, get_baselines=False, 
                 initialWB=None, celsius=None):
        """
        Sets up and checks new Block object.

        - Check that profiles list contains a list of three (3) profiles
        - Generate list of initial profiles
        - Generate list of profile directions
        - Verify three profile directions are orthogonal (['a', 'b', 'c'])
        - Generate list of ray paths
        - Verify three ray path directions are compatible with directions list
        """
        if len(profiles) != 3:
            print('keyword profiles must be a list of 3 pynams profiles')
            print('If you only have 1 or 2, make a dummy profile')
            return
        
        self.profiles = profiles
        self.folder = folder
        self.name = name
        self.time_seconds = time_seconds
        self.sample = sample        
        self.peak_diffusivities = []
        self.peak_diffusivities_errors = []
        self.celsius = celsius
        
        d = []
        r = []
        ip = []
        L = []
        
        for prof in self.profiles:
            d.append(prof.direction)
            r.append(prof.raypath)
            L.append(prof.length_microns)
            prof.time_seconds = self.time_seconds

            if prof.sample is None and self.sample is not None:
                prof.sample = self.sample
            elif prof.sample is not None and self.sample is None:
                self.sample = prof.sample
            elif prof.sample != self.sample:
                print('Warning: profile sample does not match wb sample')
                
            if prof.positions_microns is None:
                print('profile', prof.profile_name)
                print('Needs positions_microns attribute')

            if prof.initial_profile is None:
                prof.initial_profile = prof
            ip.append(prof.initial_profile)
            
            if prof.spectra is None:
                prof.make_spectra()
                
            for spec in prof.spectra:
                if self.sample is not None:
                    if prof.raypath == 'a':
                        meanthick = np.mean(self.sample.length_a_microns)
                    elif prof.raypath == 'b':
                        meanthick = np.mean(self.sample.length_b_microns)
                    elif prof.raypath == 'c':
                        meanthick = np.mean(self.sample.length_c_microns)
                    try:
                        spec.thickness_microns = meanthick
                    except NameError:
                        print('Unable to determine thicknesses from sample')
                    
        self.directions = d
        self.raypaths = r
        self.initial_profiles = ip 
        self.lengths = L

        if get_peakfit is True:
            for prof in self.profiles:
                prof.get_peakfit()

        if get_baselines is True:        
            self.get_baselines()

        if initialWB is not None:
            for idx, prof in self.profiles:
                prof.initial_profile = initialWB.profiles[idx]
                
                
    def get_peakfits(self, peak_ending='-peakfit.CSV'):
        """Get peakfit information for all profiles"""
        for prof in self.profiles:
            prof.get_peakfits(peak_ending=peak_ending)


    def get_baselines(self, initial_too=True, folder=None, delim=',', 
                      baseline_ending='-baseline.CSV',
                      print_confirmation=False):
        """Get baselines for all spectra in the block"""
        if self.initial_profiles is None:
            self.setupWB()
        for prof in self.profiles:
            for spectrum in prof.spectra:
                spectrum.get_baseline(baseline_ending=baseline_ending,
                                      folder=folder, delim=delim,
                                      print_confirmation=print_confirmation)
        if initial_too is True:
            for prof in self.initial_profiles:
                for spectrum in prof.spectra:
                    spectrum.get_baseline(baseline_ending=baseline_ending,
                                      folder=folder, delim=delim, 
                                      print_confirmation=print_confirmation)

            
    def plot_showbaselines(self):
        """Plot baselines for all spectra in the block"""
        for prof in self.profiles:
            for spec in prof.spectra:
                spec.plot_showbaseline()


    def plot_subtractbaselines(self):
        """
        Plot all spectra using plot_subtractbaseline() with default settings
        """
        for prof in self.profiles:
            for spec in prof.spectra:
                spec.plot_subtractbaseline()

            
    def make_areas(self, show_plot=False, printout_area=False, peak=None):
        """Make list of areas from all profiles, including whole-block areas."""
        self.areas = []
        self.wb_areas = []
        for idx, prof in enumerate(self.profiles):
            prof.make_wholeblock()
            self.areas.append(prof.areas)
            self.wb_areas.append(prof.wb_areas)            


    def average_spectra(self):
        """
        Create and return a single spectrum that is an average of all spectra 
        in all three profiles of the Block
        """
        absorbances_per_cm = []
        for prof in self.profiles:
            profile_average_spectrum = prof.average_spectra()
            absorb = profile_average_spectrum.abs_full_cm
            absorbances_per_cm.append(absorb)
        ave_abs = np.mean(absorbances_per_cm, axis=0)
        
        avespec = Spectrum(folder=None, fname=None)
        avespec.abs_full_cm = ave_abs
        avespec.wn_full = prof.spectra[0].wn_full
        avespec.abs_raw = ave_abs
        
        if self.name is not None:
            avespec.fname = (self.name + '\naverage across all profiles')
        else:
            avespec.fname = 'average across all profiles'
            
        return avespec


    def plot_spectra(self, profile_idx=None, show_baseline=True, 
                     show_initial_ave=True,
                     show_final_ave=True, plot_all=False, 
                     initial_and_final_together=False, style=styles.style_spectrum, 
                     stylei=styles.style_initial, wn=None):
        """Plot all spectra in all or specified profile in the Block"""
        if profile_idx is None:
            proflist = ([self.profiles[0]] + 
                        [self.profiles[1]] + 
                        [self.profiles[2]])
        else:
            proflist = [self.profiles[profile_idx]]
            
        for prof in proflist:
            print(prof.profile_name)
            prof.plot_spectra(show_baseline=show_baseline, 
                  show_initial_ave=show_initial_ave, plot_all=plot_all,
                  show_final_ave=show_final_ave,
                  initial_and_final_together=initial_and_final_together,
                  style=style, stylei=stylei, wn=None)


    def plot_3panels_ave_spectra(self, peak_idx=None, peakwn=None, 
                                 top=1., high=4000, low=3000,
                                style=styles.style_spectrum_red, 
                                stylei=styles.style_initial, show_raypaths=False,
                                figsize=(6., 4), show_initial=True,
                                legloc=5, label='Final', figax3=None):
        """
        Create three suplots showing average initial and final spectra in each
        direction
        """
        if self.initial_profiles is None:
            self.setupWB()

        if figax3 is None:
            f, ax3 = plt.subplots(1,3)
            f.set_size_inches(figsize)
            for idx, ax in enumerate(ax3[:3]):
                ax.set_xlim(high, low)
                ax.set_ylim(0., top)
                if show_raypaths is True:
                    raypathstring = ''.join(('profile || ', self.profiles[idx].direction, 
                                             '\nray path || ', self.profiles[idx].raypath))
                    ax.text(3500, top-top*0.22, raypathstring, 
                            backgroundcolor='w', horizontalalignment='center')
            
        else:
            ax3 = figax3
            
        for k in range(3):
            prof = self.profiles[k]
            avespec = prof.average_spectra()
            ax3[k].plot(avespec.wn_full, avespec.abs_full_cm, label=label, 
                        **style)
        
            if peak_idx is not None:
                if self.profiles[k].peakpos is None:
                    self.profiles[k].get_peakfit()
                peakpos = self.profiles[k].peakpos
                peakwn = peakpos[peak_idx]
                ax3[k].plot([peakwn, peakwn], [0, top], color='r')

            if show_initial is True:
                iprof = self.initial_profiles[k]
                if iprof is not None:
                    initspec = iprof.average_spectra()
                    ax3[k].plot(initspec.wn_full, initspec.abs_full_cm, 
                            **stylei)

        plt.setp(ax3[1].get_yticklabels(), visible=False)
        plt.setp(ax3[2].get_yticklabels(), visible=False)
        ax3[1].set_xlabel('wavenumber (cm$^{-1}$)')
        ax3[0].set_ylabel('absorbance (cm$^{-1}$)')
        

        if show_initial is True:
            ax3[1].legend(loc=legloc)
        if figax3 is None:
            plt.tight_layout()
            plt.gcf().autofmt_xdate()
            tit = ' '.join(('Averaged profiles for', self.name))
            if peak_idx is not None:
                tit = str.join(tit, ', peak at ', str(peakpos[peak_idx]), '/cm')
            ax3[1].set_title(tit, zorder=100) # must come after tight_layout

        plt.subplots_adjust(top=0.85, bottom=0.25)
        if figax3 is None:
            return f, ax3


    def xy_picker(self, peak_idx=None, wholeblock=True, heights_instead=False,
                  centered=True, unit='microns'):
        """ Pick out and return x and y-data for 3D plotting and diffusion """
        positions = []
        y = []
            
        for prof in self.profiles:
            positions.append(prof.positions_microns)

            # Bulk hydrogen            
            if peak_idx is None:
                
                # whole-block
                if wholeblock is True:
                    try:
                        y_to_add = prof.wb_areas
                    except AttributeError:
                        prof.make_wholeblock(peakfit=False, show_plot=False)
                        y_to_add = prof.wb_areas
                
                # absolute areas
                else:
                    try:
                        y_to_add = prof.areas
                    except AttributeError:
                        prof.make_areas()
          
            # Peak-specific                
            else:
                if heights_instead is True:
                    try:
                        y_to_add = prof.peak_heights[peak_idx]
                    except AttributeError:
                        try:
                            for idx, spec in enumerate(prof.spectra):
                                h = spec.peak_heights[peak_idx]
                                prof.peak_heights[peak_idx][idx] = h
                        except AttributeError:
                            print('Need profile.peak_heights')
                            return
                else:
                    try:
                        y_to_add = prof.peak_areas[peak_idx]
                    except AttributeError:
                        try:
                            for idx, spec in enumerate(prof.spectra):
                                a = spec.peak_areas[peak_idx]
                                prof.peak_areas[peak_idx][idx] = a
                        except AttributeError:
                            print('Need profile.peak_areas')
                            return

                if wholeblock is True:
                    peak_wb, peakwn = prof.get_peak_wb_areas(peak_idx, 
                                           heights_instead=heights_instead)
                    y_to_add = peak_wb
                        
            y.append(y_to_add)
            
        if centered is True:
            a = np.mean(self.profiles[0].sample.length_a_microns) / 2.
            b = np.mean(self.profiles[1].sample.length_b_microns) / 2.
            c = np.mean(self.profiles[2].sample.length_c_microns) / 2.
            halflengths = [a, b, c]
            for idx in range(3):
                positions[idx] = positions[idx] - halflengths[idx]
            
        return positions, y


    def plot_areas_3panels(self, peak_idx=None, 
                           axes3=None, 
                           centered=False,
                           ytop=None, 
                           heights_instead=False, 
                           wholeblock=False,
                           xerror=0., 
                           yerror=None, 
                           scale=1., 
                           pie=False,
                           styles3=[styles.style_points]*3,
                           unit='microns',
                           show_line_at_1=False, 
                           show_data=True,
                           show_errorbars=True):
        """
        Plots areas (default) or ratio of area to initial area 
        (wholeblock=True) for all 3 profiles.  
        
        Returns figure handle and a list of 3 axes handles (default) unless
        axes3 is not equal to None. 
        
        Additional keywords are similar to those for profile.plot_areas
        """
        if peak_idx is not None:
            prof = self.profiles[0]
            spec = prof.spectra[0]
            try:
                peakpos = self.peakpos
            except AttributeError:
                try:
                    peakpos = prof.peakpos
                except AttributeError:
                    try:
                        peakpos = spec.peakpos
                    except AttributeError:
                        print('Need peak positions in peakpos attribute')
                        return

        positions, y = self.xy_picker(peak_idx=peak_idx, 
                                      wholeblock=wholeblock,
                                      heights_instead=heights_instead, 
                                      centered=centered, unit=unit)
        
        if peak_idx is not None:
            tit = ' '.join(('Peak at', str(peakpos[peak_idx]), '/cm'))
        else:
            tit = 'Bulk hydrogen'

        if ytop is None:
            z = []
            for ynum in y:
                if len(ynum) > 0:
                    z.append(max(ynum))
            ytop = max(z) + 0.1*max(z)

        if unit == 'microns':
            lengths = self.lengths
        elif unit == 'mm':
            lengths = np.array(self.lengths) / 1000.
        else:
            print('unit must be microns (default) or mm')
            return
                
        if show_data is False:
            positions = [[], [], []]
            y = [[], [], []]
            
        # Sent positions and areas to plotting command
        if axes3 is not None:
            styles.plot_3panels(positions, y, lengths, figaxis3=axes3,
                                styles3=styles3, ytop=ytop, 
                                wholeblock=wholeblock,
                                show_line_at_1=show_line_at_1,
                                heights_instead=heights_instead,
                                use_errorbar=show_errorbars,
                                yerror=yerror, unit=unit, scale=scale,
                                xerror=xerror, centered=centered)
            axes3[1].set_title(tit)                                
        else:
            fig, ax = styles.plot_3panels(positions, y, lengths,
                                          styles3=styles3, ytop=ytop, 
                                          wholeblock=wholeblock,
                                          show_line_at_1=show_line_at_1,
                                          heights_instead=heights_instead,
                                          use_errorbar=show_errorbars,
                                          yerror=yerror, unit=unit,
                                          xerror=xerror, centered=centered,
                                          scale=scale)
            ax[1].set_title(tit)
            fig.set_size_inches(6.5, 3.)
            fig.autofmt_xdate()

        # add pie chart showing % of total height or area
        if pie is True:
            if peak_idx is None:
                pass
            else:
                ax_pie = fig.add_subplot(339)
                atot, htot = prof.get_area_total()
                
                if heights_instead is False:
                    a = spec.peak_areas[peak_idx] / atot
                    size = [a, 1. - a]
                    tit = '% total area'
        
                else:
                    h = spec.peak_heights[peak_idx] / htot
                    size = [h, 1. - h]
                    tit = '% sum of heights'
            
                colors = ['k', 'w']
                plt.pie(size, colors=colors, 
                        startangle=90,
                        radius=0.25, center=(0, 0), frame=False)
                ax_pie.axis('equal')
                ax_pie.set_title(tit)

        if axes3 is None:
            return fig, ax


    def make_composite_peak(self, peak_idx_list):
        """Make composite peaks for all spectra in the Block"""
        for prof in self.profiles:
            for spec in prof.spectra:
                spec.make_composite_peak(peak_idx_list)
            prof.get_peak_info()


    def print_spectra_names(self, show_initials=True):
        """Print out fnames of all spectra in the Block."""
        if show_initials is True:
            if self.initial_profiles is None:
                self.setupWB()
                
            if self.initial_profiles is not None:
                print('--Initial profiles--')
                for prof in self.initial_profiles:
                    print(prof.profile_name)
                    spec_list = []
                    for spectrum in prof.spectra:
                        spec_list.append(spectrum.fname)
                    print(spec_list)
                    print(' ')
            else:
                print('No initial profiles given')

        if self.profiles is not None:

            print('--Final profiles--')
            for prof in self.profiles:
                print(prof.profile_name)
                spec_list = []
                for spectrum in prof.spectra:
                    spec_list.append(spectrum.fname)
                print(spec_list)
                print(' ')


    def make_baselines(self,
                      raw_data=False, 
                      wn_low=3200, 
                      wn_high=3700, 
                      linetype='line', 
                      spline_kind='cubic', 
                      spline_wn_low=3000,
                      spline_wn_high=4000,
                      curvature=None, 
                      force_through_wn=None,
                      polynomial_order=None,
                      show_fit_values=False, 
                      show_plot=False,
                      abs_high=None, 
                      abs_low=None,
                      abs_smear_high=0, 
                      abs_smear_low=0,
                      store_baseline=True
                      ):
        """
        Make spectra baselines for all spectra the Block. 
        
        Keywords are similar to spectrum.make_baseline()
        """  
        for prof in self.profiles:
            prof.make_baselines(raw_data=raw_data, wn_low=wn_low, 
                               wn_high=wn_high, linetype=linetype, 
                               spline_kind=spline_kind,
                               spline_wn_high=spline_wn_high,
                               spline_wn_low=spline_wn_low,                               
                               curvature=curvature,
                               force_through_wn=force_through_wn,
                               polynomial_order=polynomial_order,
                               show_fit_values=show_fit_values,
                               show_plot=show_plot, abs_high=abs_high,
                               abs_low=abs_low, 
                               abs_smear_high=abs_smear_high,
                               abs_smear_low=abs_smear_low,
                               store_baseline=store_baseline)


    def save_baselines(self, initial_too=True, 
                       baseline_ending='-baseline.CSV'):
        """Make and save spectra baselines for all spectra in the Block."""
        for prof in self.profiles:
            for spectrum in prof.spectra:
                spectrum.save_baseline(baseline_ending=baseline_ending)


    def plot_peakfits(self, initial_too=False, profile_idx=None, legloc=1):
        """ Plot peakfits for all spectra in all profiles in the Block """
        if profile_idx is None:
            for prof in self.profiles:
                prof.plot_peakfits(initial_too, legloc=legloc)
        else: 
            self.profiles[profile_idx].plot_peakfits(initial_too, 
                                                    legloc=legloc)


    def print_max_arearatio(self, peak_idx=None, heights_instead=False):
        """ Prints out the maximum whole-block area ratio observed 
        in any profile of the Block for specified peak_idx"""
        if peak_idx is None:
            self.setupWB(False, True)
        else:
            self.get_peakfit()
            self.setupWB(True, False)

        a = []

        for prof in self.profiles:
            if peak_idx is None:
                maxval = max(prof.wb_areas)
            else:
                prof.make_wholeblock(peakfit=True)
                if heights_instead is False:
                    maxval = max(prof.peak_wb_areas[peak_idx])
                else:
                    maxval = max(prof.peak_wb_heights[peak_idx])
            a.append(maxval)
        print('\n', self.name)
        print(max(a))
        
        return max(a)


    def print_peakfits_ave(self, printall=True, print_max=False, 
                           print_aves=True):
        """ Prints out and returns average peak areas, peak heights, and 
        sum of the average peak areas summed over all profiles"""
        areas = []
        heights = []
        totalareas = []
        max_a = []
        max_h = []
        for prof in self.profiles:
            a, h, totala, ma, mh = prof.print_peakfits_ave(printout=printall)
            areas.append(a)
            heights.append(h)
            totalareas.append(totala)
            max_a.append(ma)
            max_h.append(mh)

            if print_max is True:
                print(ma)
                print(mh)
            
        if print_aves is True:
            asum = np.array(np.sum(areas, axis=0))
            hsum = np.array(np.sum(heights, axis=0))
            tasum = np.sum(totalareas)
    
            print('\n', self.name)
            print('peak positions (cm-1)')
            print(self.profiles[0].spectra[0].peakpos)
            print('\naverage peak areas summed over all profiles (cm-2)')
            print(asum)
            print('\naverage peak heights summed over all profiles (cm-1)')
            print(hsum)
            print('\naverage total area summed over all profiles (cm-1)')
            print(tasum)
        return asum, hsum, tasum


    def print_diffusivities(self, peak_idx=None, profile_idx=None,
                            show_plot=False, top=1.5):
        """Print diffusivities for each profile"""
        if show_plot is True:
            self.plot_3panels_ave_spectra(peak_idx=peak_idx, top=top)
                     
        if peak_idx is None and profile_idx is None:
            for prof in self.profiles:
                prof.get_diffusivities()
                prof.print_diffusivities()
            return
        
        if profile_idx is None:
            for prof in self.profiles:
                k = peak_idx
                print('\n', prof.profile_name)
                print('peak position and log10(diffusivity in m2/s)')
                print('bulk H :', prof.D_area_wb, '+/-', \
                    prof.D_area_wb_error)
                print(prof.peakpos[k], ':', prof.D_peakarea_wb[k],\
                    '+/-', prof.peak_D_area_wb_error[k])
            return
            
        prof = self.profiles[profile_idx]
        print('\n', prof.profile_name)
        print('peak positions, log10(D in m2/s), D errors')
        print('bulk H :  ', prof.D_area_wb, '+/-', \
                prof.D_area_wb_error)
        print(prof.peakpos)
        print(prof.D_peakarea_wb)
        print(prof.peak_D_area_wb_error)


    def save_diffusivities(self, folder=None, 
                           file_ending='-diffusivities.txt'):
        """Save diffusivities for all profiles in Block to files"""
        for prof in self.profiles:
            prof.save_diffusivities(folder, file_ending)

            
    def get_diffusivities(self, folder=None, 
                           file_ending='-diffusivities.txt'):
        """Gets diffusivities for all profiles in Block from saved files"""
        if folder is None:
            folder = self.folder
        for prof in self.profiles:
            prof.get_diffusivities(folder, file_ending)

    def diffusion_profiles(self, 
                           init=1., 
                           fin=0, 
                           wholeblock_data=False,
                           wholeblock_diffusion=False,
                           peak_idx=None, 
                           time_seconds=None, 
                           log10D_m2s=[-12., -12., -12.], 
                           erf_or_sum='erf', 
                           points=50, 
                           heights_instead=False, 
                           approximation1D=False):
        """
		Calculate diffusion profiles in the Block

        Requires:
            * time in seconds either explicitly passed here or as an
        	  attribute of the Block object
            * list of log10 diffusivities in m2/s 
              (default=[-12, -12, -12])
            * initial value (default init=1)
            * final value (default fin=0)
        
        Defaults:
        	* Calculates 3D non-path-integrated profiles 
        	  (wholeblock_diffusion=False) as opposed to path-integrated
        	  whole-block profiles (wholeblock_diffusion=True)
        	* Assumes the data are not normalized to any initial 
        	  value (wholeblock_data=False) as opposed to normalized
        	  (wholeblock_data=True). This matters for determining the 
        	  maximum and minimum values

        Returns lmfit parameters, x-data, and y-data for 3-dimensional 
        diffusion in a block.        
        """
        if self.lengths is None:
            self.setupWB(peakfit=False, make_wb_areas=False)
        if self.lengths is None:
            print('Need to setup self.lengths, which is in microns')
            return False

        if self.directions is None:           
            self.setupWB(peakfit=False, make_wb_areas=False)

        if self.initial_profiles is None:
            self.setupWB(peakfit=False, make_wb_areas=False)

        if self.raypaths is None:
            self.setupWB(peakfit=False, make_wb_areas=False)

        if time_seconds is None:
            if self.time_seconds is not None:
                time_seconds = self.time_seconds
            else:
                print('Need time information')
                return False, False, False
                
        # Pick which diffusivities to use
        if log10D_m2s is not None:
            D3 = models.D_checker(log10D_m2s)
        elif wholeblock_diffusion is True and peak_idx is None:
            D3 = self.D_area_wb
        else:
            D3 = []
            for prof in self.profiles:
                D = prof.D_picker(wholeblock_data, heights_instead, peak_idx)
                D3.append(D)
        if D3 is None or 0.0 in D3:
            print('D3:', D3)
            print('\nNeed diffusivities.')
            print('Input directly as diffusivities_log10D_m2s')
            print('or input bulk in profile.D_area_wb')
            print('or peak_diffusivities at specified peak_idx\n')
            return False

        L3 = self.lengths
        
        # need initial and final as unit values, so max=1
        # compare values with maximum
        if wholeblock_data is True:
            maxval = max([init, fin, 1.])
        else:
            maxval = max([init, fin])
        init_unit = init/maxval
        fin_unit = fin/maxval
        
        params = models.params_setup3D(L3, D3, time_seconds, 
                                       init_unit, fin_unit)

        if wholeblock_diffusion is True:
            xdiff, ydiff = models.diffusion3Dwb_params(params, 
                                                   raypaths=self.raypaths, 
                                                   erf_or_sum=erf_or_sum,
                                                   show_plot=False)           
        else:
            v, ydiff, xdiff = models.diffusion3Dnpi_params(params, 
                                                           points=points, 
                                                           centered=False)
        
        if wholeblock_data is False:
            maxareas = []
            for prof in self.profiles:
                if len(prof.fnames) > 0:
                    try:
                        maxa = np.max(prof.areas)
                    except AttributeError:
                        prof.make_areas()
                        maxa = np.max(prof.areas)
                    maxareas.append(maxa)
            ydiff = np.array(ydiff) * np.max(maxareas)

        return params, xdiff, list(ydiff)

            
    def plot_diffusion(self, wholeblock_diffusion=False, 
    				   wholeblock_data=True, 
                       peak_idx=None, 
                       time_seconds=None, 
                       log10D_m2s=[-12., -12., -12.], 
                       init=1., 
                       fin=0, 
                       erf_or_sum='erf', 
                       show_plot=True, 
                       show_data=True,
                       xaxis='centered',
                       show_slice=False, 
                       label4legend=[None, None, None],
                       axes3=None, 
                       points=50, 
                       top_spectra=1.0,
                       ytop=None, 
                       numformat='{:.1f}', 
                       heights_instead=False, 
                       centered=True,
                       approximation1D=False, 
                       labelD=True,
                       show_errorbars=True, 
                       labelDy=None, 
                       labelDx=[None, None, None],
                       style_data=styles.style_points,
                       style_diffusion=styles.style_1,
                       show_line_at_1=True,
                       ):
        """
        Plot 3D diffusion profiles for Block.

        Applies 3-dimensionsal diffusion equations in 
        pynams.diffusion.models and plots non-path-integrated 3-dimensional
        diffusion curves (default) or path-integrated whole-block profiles
        described in Ferriss et al. 2015 (wholeblock_diffusion=True).
        
        If show_data is True (the default), also plots the data, either
        directly as the measured areas (default) or peak heights
        (with heights_instead=True) or as the ratio of the
        measured area to a best-fit line through the initial areas
        (wholeblock_data=True)
        
        Default initial value is 1, and default final value is 0. 
        Change keywords init and fin to change these values, including 
        to switch from diffusion out to diffusion in.
        
        If axes3 = a list of 3 axes handles, the data and diffusion curve
        are plotted there. Otherwise, the figure handle and a list of 
        the 3 axes handles are returned.

        The diffusivities are automatically labeled on the plot. 
            To remove the labels, set labelD=False
            To change the number of digits, play with numformat keyword.
            To change the y position across all axes, use labelDy.
            To change the x positions, pass *a list of x-values* into labelDx.
        
        Change the maximum y value with the top keyword.
        """        
        D3 = models.D_checker(log10D_m2s)
        hinstead = heights_instead
        approx = approximation1D
        params, xdiff, ydiff = self.diffusion_profiles(wholeblock_data=wholeblock_data,
                                                   peak_idx=peak_idx,
                                                   time_seconds=time_seconds,
                                                   log10D_m2s=D3,
                                                   erf_or_sum=erf_or_sum,
                                                   wholeblock_diffusion=wholeblock_diffusion,
                                                   points=points,
                                                   heights_instead=hinstead,
                                                   init=init, fin=fin,
                                                   approximation1D=approx)

        if params is False:
            return False, False

        if axes3 is None:
            fig, axes3 = self.plot_areas_3panels(peak_idx=peak_idx, ytop=ytop,
                                              wholeblock=wholeblock_data, 
                                              heights_instead=heights_instead,
                                              show_line_at_1=False,
                                              label4legend=label4legend,
                                              styles3=[style_data]*3,
                                              centered=centered,
                                              show_errorbars=show_errorbars)
        if centered is True:
            for idx_len in range(3):
                xdiff[idx_len] = xdiff[idx_len] - (self.lengths[idx_len]/2.)

        if wholeblock_data is False:
            maxareas = []
            for prof in self.profiles:
                if len(prof.fnames) > 0:
                    try:
                        maxa = np.max(prof.areas)
                    except AttributeError:
                        prof.make_areas()
                        maxa = np.max(prof.areas)
                    maxareas.append(maxa)
            init = np.max(maxareas)

        styles.plot_3panels(xdiff, np.array(ydiff), 
                            show_line_at_1=show_line_at_1, 
                            figaxis3=axes3, init=init, ytop=ytop,
                            styles3=[style_diffusion]*3, 
                            centered=centered)

        # label diffusivities
        if labelD is True:
            for k in range(3):
                if centered is True:
                    labelDx[k] = 0.
                else:
                    labelDx[k] = self.lengths[k]/2.
                dlabel = ''.join(('logD ', str(numformat.format(D3[k])), 
                                  ' m$^2$/s'))
                if labelDy is None:
                    ytop = axes3[0].get_ylim()[1]
                    labelDy = ytop - ytop*0.2
                axes3[k].text(labelDx[k], labelDy, dlabel, 
                              horizontalalignment='center',
                              verticalalignment='center',
                              backgroundcolor='w')
   
        try:
            return fig, axes3
        except NameError:
            pass

       
    def fitD(self, peak_idx=None, 
    		 init=1., 
    		 fin=0.,
             log10Ds_m2s=[-13., -13., -13.], 
             heights_instead=False, 
             wholeblock_data=True,
             vary_initial=False, 
             vary_final=False, 
             vary_diffusivities=[True, True, True],
             erf_or_sum='erf', 
             show_plot=True, 
             wholeblock_diffusion=True, 
             centered=True, 
             show_initial_guess=True, 
             style_initial=None,
             style_final={'color':'red'}, 
             points=50, 
             top=1.2):
        """
        Fit 3D diffusion curves to Block data.
        
        You can solve for any or all of the following:
            * diffusivities in all three directions (vary_diffusivities;
              default=True)
            * the final value (vary_final; default=False)
            * the initial value (vary_initial; default=False)
            
        Returns: initial, final, list of diffusivities.
        
        You can model using path integration (wholeblock_diffusion=True by
        default) or non-path-integrated 3D diffusion.
        
        You can input your own starting guesses for the diffusivities in
        a list passed into log10D_m2s. 
        
        The data can be either "wholeblock", which has a range from 0 to 1, 
        (default) or not (set wholeblock_data=False)
        
        To hold one or more diffusivities constant: pass lists of 
        diffusivities in through log10Ds_m2s (default=[-13, -13, -13]) 
        and what to hold constant (vary_diffusivities=[True, True, True]
        by default) 
        """        

        # x and y are the data that we will fit to, centered for fitting
        try:
            x, y = self.xy_picker(peak_idx=peak_idx, 
                                  wholeblock=wholeblock_data, 
                                  heights_instead=heights_instead, 
                                  centered=True)
        except AttributeError:
            print('problem in block.xy_picker getting x and y data')
            return
        
        if wholeblock_data is False:
            maxy = max([max(areas) for areas in y])
            y = np.array([np.array(areas) for areas in y]) / maxy
            init_unit = init / maxy
            fin_unit = fin / maxy
        else:
            init_unit = init
            fin_unit = fin


        # set up fitting parameters in the required format
        params = models.params_setup3D(microns3=self.lengths, 
                 log10D3=log10Ds_m2s,
                 time_seconds=self.time_seconds, 
                 initial=init_unit, final=fin_unit,
                 vinit=vary_initial, vfin=vary_final,
                 vD=vary_diffusivities)

        # other keywords needed for forward model
        dict_fitting = {'points' : points,
                        'erf_or_sum' : erf_or_sum} 

        if wholeblock_diffusion is True:
            # need raypaths and don't plot twice
            if self.raypaths is None:
                self.setupWB()
            dict_fitting['raypaths'] = self.raypaths
            dict_fitting['show_plot'] = False
            
            x = np.array(x)
            y = np.array(y)
            # run the minimizer
            lmfit.minimize(models.diffusion3Dwb_params, 
                           params, args=(x, y), 
                           kws=dict_fitting
                           )


    def invert(self, grid_xyz, symmetry_constraint=True, 
               smoothness_constraint=True, rim_constraint=True, 
               rim_value=None, weighting_factor_lambda=0.2, 
               show_residuals_plot=True):
        """
		Invert Block data to estimate internal concentrations. Tomography!

        Takes a list of three whole-block concentration profiles (either A/Ao 
        or water ok but must be consistent for all three) in three orthogonal 
        directions and list of three integers to indicate number of divisions
        in each direction. Returns matrix of values in each grid cell. 
        Default plot showing residuals for how well results match the 
        whole-block observations. 

        Ferriss did this in MATLAB for the 2015 paper establishing
        the whole-block method but never needed it for the papers on 
        clinopyroxene and olivine, so this function never got written 
        in python.
        """
        pass


    def make_peakheights(self, peaks=[3600, 3525, 3356, 3236]):
        """
		Estimate peak heights for all spectra in Block. 

        Requires a list of peak wavenumber locations in cm-1
            (default peaks=[3600, 3525, 3356, 3236])
        Creates or overwrites profiles peak positions and peak_heights using 
        that baseline, not heights from gaussian curves
        """
        for prof in self.profiles:
            prof.peakpos = peaks
            prof.peak_heights = []
            for peak in (peaks):
                heights = []
                for spec in prof.spectra:
                    idx = np.abs(peak - spec.base_wn).argmin()
                    height_base = spec.base_abs[idx]
                    idx = np.abs(peak - spec.wn_full).argmin()
                    height_abs = spec.abs_full_cm[idx]                        
                    height = height_abs - height_base
                    heights.append(height)
                prof.peak_heights.append(heights)


"""
Code for processing and plotting FTIR spectra.

The central concept is an object class called Spectrum, which contains
relevant attributes and methods.

A few defaults are set up that are geared toward H in nominally anhydrous
minerals (NAMs) such a plotting wavenumber range from 3000 to 4000 /cm and
creating baselines between 3200 and 3700 /cm. 

@author Elizabeth Ferriss

"""
from __future__ import print_function, division, absolute_import
import numpy as np
import os
import matplotlib.image as mpimg
import matplotlib.pyplot as plt
import pynams.styles as styles
import scipy.interpolate as interp
import pandas as pd
from uncertainties import ufloat
from . import pynams
from mpl_toolkits.axes_grid1.parasite_axes import SubplotHost
import gc
from scipy import signal as scipysignal

class Spectrum():
    """
    For plotting and manipulating Fourier transform infrared (FTIR) spectra.
    
    Requires an fname, which is the first part of the FTIR filename, e.g.,
    if your raw FTIR spectrum is saved in C:my_folder as my_file.CSV,
    the fname is 'my_file' and the filetype='.CSV'. The folder should be 
    set to 'C:my_folder'. The fname is critical both for locating your file
    and importing your data and because it is used as the default heading and 
    label for a lot of the plotting functions, so it is required. 
    The simplest call would look like this:
        my_spectrum = Spectrum('test_spectrum')
    
    The folder defaults to the present working directory, but you'll 
    probably need to set the folder explicitly using a string like this one:
        'C:\\Users\\Ferriss\\Documents\\Code\\olivine\\'

    The filetype defaults to .CSV, and that's the filetype this code handles
    best, although it's done ok with .txt. 
        
    Once you make your Spectrum, it will automatically open the file and read 
    in the wavenumbers and absorbances. There is a test file test.CSV in 
    the pynams folder.
    
    Use the bound method plot_spectrum() to plot the spectrum, e.g.,
    my_spectrum.plot_spectrum()
    
    The sample thickness can be input in the following ways: 
        1. You can set thickness in microns directly as a keyword when you
           make the Spectrum:
               Spectrum(fname, thickness_microns=x)
               
        2. You can specify a sample, a pynams Sample object, 
           that has thickness information input as thickness_thinslab_microns:
               Spectrum(fname, sample=my_sample)
               
        3. By specifying a sample, a pynams Sample object, that has 
           length information in at least one direction AND specifying the 
           raypath as 'a', 'b', or 'c' or the equivalent 'x', 'y', or 'z'.
           The thickness is then set the as the length of the Sample parallel 
           to the raypath. If the raypath is None (the default) or 'u', it
           acts like Example #2 above in using whatever is set as the sample's
           thickness_thinslab_microns.
           
        4. You can have it guess the thickness based on the Si-O overtones,
           by calling the bound method get_thickness_from_SiO():
               my_spectrum.get_thickness_from_SiO()
    
    The baseline is by default a line between wavenumbers 3200 cm-1 
    (base_low_wn) and 3700 (base_high_wn), and a quadratic baseline would
    deviate from that line relative to base_mid_wn in the middle. Those 
    values can be most easily set under make_baseline for an individual
    spectra, but for handling whole profiles at a time, it's convenient
    to be able to set these attributes and then be able to refer to them
    later on.
    """
    def __init__(self, fname, filetype='.CSV', folder='', sample=None,
                 thickness_microns=None, raypath=None, polar=None, 
                 base_low_wn=3200, base_high_wn=3700,
                 base_mid_wn=3550):
        """
        Automatically locates the FTIR file and pulls out the wavenumber 
        (wn_full) and raw absorbance data (abs_raw) and sets the 
        thickness_microns attribute if it input directly or as part of a 
        sample. 
        """
        self.fname = fname
        self.folder = folder
        self.sample = sample
        self.base_high_wn = base_high_wn
        self.base_low_wn = base_low_wn
        self.base_mid_wn = base_mid_wn
        self.polar = polar 
        self.filetype = filetype
        self.raypath = raypath        
        
        if self.fname is None:
            return
        
        self.filename = self.folder + self.fname + self.filetype

        if os.path.isfile(self.filename):
            if self.filetype == '.CSV':
                signal = pd.read_csv(self.filename, header=None)
            elif self.filetype == '.txt':
                try:
                    signal = np.loadtxt(self.filename, delimiter='\t', 
                                        dtype=None) 
                except ValueError:
                    print('\nProblem reading this file format. Try .CSV')
            else:
                print('For now only CSV and txt files work.')

            signal = signal.sort_values(by=[0])
            self.wn_full = np.array(signal[0])
            self.abs_raw = np.array(signal[1])
            
            if thickness_microns is not None:
                self.thickness_microns = thickness_microns
            else:
                if sample is not None:
                    idx = styles.get_iorient(raypath)
                    self.thickness_microns = sample.thickness_microns[idx]
                else:
                    self.thickness_microns = None
        else:
            print('There is a problem finding the file.')
            print('filename =', self.filename)
            print('Maybe check the folder name')
            

    def plot_spectrum(self, axes=None, style=None, offset=0., 
                      label=None, wn_xlim_left=4000., wn_xlim_right=3000., 
                      pad_top=0.1, pad_bot=0., plot_raw=False):
        """
        Produces a plot of the FTIR spectrum, and returns the figure and axes 
        handles. If you don't know what those are, check out this site: 
            http://matplotlib.org/faq/usage_faq.html
        
        If you already have an axis that you want to add a plot onto, set 
        axes=the handle for the pre-existing axes object, e.g.,
            fig, ax = my_spectrum.plot_spectrum()
            my_second_spectrum.plot_spectrum(axes=ax)
        That should produce a single figure with both my_spectrum and 
        my_second_spectrum plotted on top of each other.
        
        The style of the line can be changed using the keyword style and
        passing in a dictionary. There are several preset style dictionaries
        included in pynams.styles, so, for instance, try:
            from pynams import styles
            my_spectrum.plot_spectrum(style=styles.style_lightgreen)
        Note that the axis takes on the x- and y-limits from the last 
        spectrum plotted, and you can change those after plotting using
        ax.set_xlim() and ax.set_ylim(). 
        
        If you want to move a spectrum up or down, use the offset keyword.
        
        If you add a legend to your axes object with ax.legend(), the label
        given to your spectrum will default to the spectrum's fname. If you
        want to use a different label, pass a 'string' for the label keyword.
        
        The x and y axis limits are set by the keywords wn_xlim_left,
        wn_xlim_right, pad_top, and pad_bot, which default to showing the
        4000-3000 cm-1 wavenumber range where O-H stretching peaks show
        up in nominally anhydrous minerals and zoomed in on the spectrum.
        
        The keyword plot_raw clarifies whether or not to plot the raw
        absorbance data or absobance data that has been divided by 
        thickness and offset such that the lowest point between wavenumbers
        4000 and 3000 cm-1 is set equal to 0. The default is plot_raw=False,
        meaning that the thickness-divided-and-offset spectrum will be plotted,
        but if not thickness data are available, then it will revert to
        plotting just the raw absorbance data.
        """
        if plot_raw is True:
            absorbance = self.abs_raw
        else:
            try:
                absorbance = self.abs_full_cm
            except AttributeError:
                self.divide_by_thickness()
                self.start_at_zero()
                try:
                    absorbance = self.abs_full_cm
                except AttributeError:
                    absorbance = self.abs_raw
                    plot_raw = True
                
        if axes is None:
            fig, ax = styles.plot_spectrum_outline(wn_xlim_left=wn_xlim_left,
                                                   wn_xlim_right=wn_xlim_right)
            fig.set_size_inches(6, 6)

        else:
            fig = None
            ax = axes

        if plot_raw is True:
            ax.set_ylabel('raw absorbance')
        else:
            self.start_at_zero()
            
        if label is None:
            label = self.fname

        if style is not None:            
            style_to_use = style.copy()
            style_to_use.update({'label' : label})
            ax.plot(self.wn_full, absorbance + offset, **style_to_use)
        else:
            ax.plot(self.wn_full, absorbance + offset, label=label)
        
        ax.set_xlim(wn_xlim_left, wn_xlim_right)
        
        ylow, yhigh = styles.ylim_picker(self, wn_xlim_left=wn_xlim_left,
                                         wn_xlim_right=wn_xlim_right, 
                                         pad_top=pad_top, pad_bot=pad_bot)
        ax.set_ylim(ylow, yhigh)
        ax.set_title(self.fname)
        return fig, ax
    

    def orientation(self, label=None):
        """
        Returns a figure comparing your spectrum in the wavenumber range
        2200-1200 with spectra for oriented olivine with polarized radation
        in different directions to help guess the olivine orientation.
        
        Use label keyword to change the title on your spectrum from the fname.
        
        """
        # get the image from Lemaire et al. 2004
        thisfolder = os.path.dirname(__file__)
        img_file = '\\'.join((thisfolder, 'Lemaire2004Figure1a.png'))
        img=mpimg.imread(img_file)
        
        # set up subplots and figure
        fig = plt.figure()
        ax1 = SubplotHost(fig, 1,2,1)
        ax2 = SubplotHost(fig, 1,2,2)
        fig.add_subplot(ax1)
        fig.add_subplot(ax2)
        
        # show the image
        plt.imshow(img)
        plt.axis('off')
        ax2.set_title('Oriented olivine\nLemaire et al. 2004')
        
        # show the spectrum with the same scale
        wn_high = 2200.
        wn_low = 1200.

        self.plot_spectrum(pad_top=0.4, wn_xlim_left=wn_high, axes=ax1,
                           wn_xlim_right=wn_low, plot_raw=True)
        
        # zoom in on Si overtone areas
        index_lo = (np.abs(self.wn_full-wn_low)).argmin()
        index_hi = (np.abs(self.wn_full-wn_high)).argmin()        
        SiO = self.abs_raw[index_lo:index_hi]
        ax1.set_ylim(min(SiO), max(SiO)+0.2*max(SiO))
        
        # set spectrum title
        if label is None:
            ax1.set_title(self.fname)
        else:
            ax1.set_title(label)

        fig.set_size_inches(8, 4)
        
        ytext = ax1.get_ylim()[1] - 0.1*ax1.get_ylim()[1]
        wns = [2035, 1925, 1840, 1785, 1670, 1600]
        for idx, wn in enumerate(wns):
            ax1.plot([wn, wn], ax1.get_ylim(), '-r')
            ax1.text(wn, ytext, str(wn), rotation=90, backgroundcolor='w',
                    va='center', ha='center', fontsize=12)
        
        ax1.set_xlabel('Wavenumber (cm$^{-1}$)')
        plt.tight_layout()
        plt.subplots_adjust(wspace=-0.1)
        return fig


    def get_thickness_from_SiO(self, show_plot=False, printout=False,
                               accept_thickness=True):
        """
        Estimates the sample thickness based on area of the Si-O overtones
        Using Eq 1 of Matveev and Stachel 2007. 
        
        If show_plot is set to True, it will show (and return figure and axes 
        handlles for) a plot of the line under the Si-O overtones showing the 
        area used to estimate the thickness. 
        
        If printout is set to True, then it will also print the exact
        area under that curve.
        
        If accept_thickness=True (default), the spectrum thickness will
        be automatically changed, and the thickness-normalized absorbance
        will be unpdated. Note that this will screw up any previous baselines.
        """
        wn_low = 1625
        wn_high = 2150

        # indices for absorbance of interest
        index_lo = (np.abs(self.wn_full-wn_low)).argmin()
        index_hi = (np.abs(self.wn_full-wn_high)).argmin()        

        # raw absorbance over Si-O overtones
        SiO_overtones = self.abs_raw[index_lo:index_hi]
        
        # temporary linear baseline under Si-O overtones
        bline = self.make_baseline(wn_low=wn_low, 
                                   wn_high=wn_high, 
                                   show_plot=show_plot,
                                   raw_data=True, 
                                   store_baseline=False
                                   )
        dx = wn_high - wn_low
        dy = np.mean(SiO_overtones - bline)
        SiOarea = dx * dy
        thickness_microns = SiOarea / 0.6366
        
        if accept_thickness is True:
            self.thickness_microns = thickness_microns
            self.divide_by_thickness()
            self.start_at_zero()
        return thickness_microns


    def find_lowest_wn_over_given_range(self, wn_mid_range_high=3500., 
                                        wn_mid_range_low=3300.,
                                        relative=True):
        """
        Take a spectrum and wavenumber range (default 3300-3500 cm-1)
        and returns the wavenumber with the lowest absorbance within that range
        """
        if self.abs_full_cm is None:
            self.start_at_zero()
    
        ## find minimum relative to a linear baseline
        if relative is True:            
            self.make_baseline(linetype='line', show_plot=False)
            idx_mid_high = (np.abs(self.base_wn-wn_mid_range_high)).argmin()
            idx_mid_low = (np.abs(self.base_wn-wn_mid_range_low)).argmin()
            if idx_mid_high == idx_mid_low:
                print('basenumber range not established. Check wavenumber range')
                return False
        
            abs_nobase = self.subtract_baseline()
            mid_abs_range = abs_nobase[idx_mid_low:idx_mid_high]
            
            mid_wn_range = self.base_wn[idx_mid_low:idx_mid_high]
            idx_abs_mid = mid_abs_range.argmin()
            WN_MID = mid_wn_range[idx_abs_mid]

        else:        
            ## finds absolute minimum over range
            idx_mid_high = (np.abs(self.wn_full-wn_mid_range_high)).argmin()
            idx_mid_low = (np.abs(self.wn_full-wn_mid_range_low)).argmin()
            mid_abs_range = self.abs_full_cm[idx_mid_low:idx_mid_high]
            mid_wn_range = self.wn_full[idx_mid_low:idx_mid_high]
            idx_abs_mid = mid_abs_range.argmin()
            WN_MID = mid_wn_range[idx_abs_mid]

        return WN_MID
    

    def find_peaks(self, sensitivity=40, show_plot=True, printout=False):
        """
        Returns wavenumbers of the most prominent peaks in wavenumber range of 
        the existing baseline. 
        
        Play with the sensivitity keyword (default=40) to change how many
        peaks it picks up. Higher sensitivity gives you fewer peaks - sorry.
        """
        abs_nobase_cm = self.subtract_baseline()
        if abs_nobase_cm is None:
            return
            
        widths = np.arange(1, sensitivity)
        peaks = scipysignal.find_peaks_cwt(abs_nobase_cm, widths)
        
        if printout is True:
            print('peaks found at the following wavenumbers:')
            print(self.base_wn[peaks])
            print('change sensitivity to find more or fewer peaks')

        if show_plot is True:
            fig, ax = self.plot_subtractbaseline()
            for idx in peaks:
                wn = self.base_wn[idx]
                ax.plot([wn, wn], ax.get_ylim(), '-r', linewidth=1.5)        

        return self.base_wn[peaks]
    
        
    def make_peakheights(self, peaks=[3600, 3525, 3356, 3236]):
        """
        Requires a list of peak wavenumber locations in cm-1
            (default peaks=[3600, 3525, 3356, 3236])
        Creates or overwrites any existing peak positions and peak_heights 
        with peak heights from current baseline
        """
        self.peakpos = peaks
        self.peak_heights = []
        for peak in (peaks):
            idx = np.abs(peak - self.base_wn).argmin()
            height_base = self.base_abs[idx]
            idx = np.abs(peak - self.wn_full).argmin()
            height_abs = self.abs_full_cm[idx]                        
            height = height_abs - height_base
            self.peak_heights.append(height)

    
    def make_average_spectra(self, spectra_list, folder=None):
        """Takes list of spectra and returns average absorbance (/cm)
        to the new spectrum (self)"""       
        list_abs_to_average = []
        list_abs_full_to_average = []
        list_wn = []
        thicknesses = []
        for sp in spectra_list:
            list_abs_to_average.append(sp.abs_raw)
            try:
                list_abs_full_to_average.append(sp.abs_full_cm)
            except AttributeError:
                sp.divide_by_thickness()
                sp.start_at_zero()
                list_abs_full_to_average.append(sp.abs_full_cm)
            list_wn.append(sp.wn_full)
            thicknesses.append(sp.thickness_microns)
        self.wn_full = np.mean(list_wn, axis=0)
        self.abs_raw = np.mean(list_abs_to_average, axis=0)
        self.abs_full_cm = np.mean(list_abs_full_to_average, axis=0)
        self.thickness_microns = np.mean(thicknesses, axis=0)
        
    
    def divide_by_thickness(self):
        """
        Divide raw absorbance by thickness
        """
        # Convert from numpy.float64 to regular python float
        # or else element-wise division doesn't work.
        try:
            if isinstance(self.thickness_microns, float) is True:
                th = self.thickness_microns
            elif isinstance(self.thickness_microns, int) is True:
                th = float(self.thickness_microns)
            else:
                th = np.asscalar(self.thickness_microns)
        except AttributeError:
            print('Need a thickness estimate.')
            print('You could try Spectrum.get_thickness_from_SiO')
            return False
            
        self.abs_full_cm = self.abs_raw * 1e4 / th
        return self.abs_full_cm
    

    def start_at_zero(self, wn_xlim_left=4000., wn_xlim_right=3000.):
        """
        Divide the raw FTIR absorbance by thickness and
        shift minimum to 0 within specified wavenumber range specified
        by wn_xlim_left and _right
        """
        index_lo = (np.abs(self.wn_full-wn_xlim_right)).argmin()
        index_hi = (np.abs(self.wn_full-wn_xlim_left)).argmin()
        indices = list(range(index_lo, index_hi, 1))

        try: 
            self.abs_full_cm = self.abs_full_cm - min(self.abs_full_cm[indices])
        except AttributeError:
            self.divide_by_thickness()
            try:
                minabs = min(self.abs_full_cm[indices])
                self.abs_full_cm = self.abs_full_cm - minabs
            except AttributeError:
                return
        except ValueError:
            print('There was a problem.')
            print('index_lo at wn', wn_xlim_right, ':', index_lo)
            print('index_hi at wn', wn_xlim_left, ':', index_hi)
            return False
        return self.abs_full_cm
    

    def make_baseline(self, 
                      raw_data=False, 
                      wn_low=3200, 
                      wn_high=3700, 
                      wn_mid=3550,
                      linetype='line', 
                      spline_kind='cubic', 
                      spline_wn_low = 3000,
                      spline_wn_high = 4000,
                      curvature=None, 
                      force_through_wn=None,
                      polynomial_order=None,
                      show_fit_values=False, 
                      show_plot=False,
                      abs_high=None, 
                      abs_low=None,
                      abs_smear_high=0, 
                      abs_smear_low=0,
                      store_baseline=True
                      ):
        """
        Makes and returns a baseline under the curve of an FTIR spectrum. 
        
        The spectrum can be either the raw absorbance (raw_data=True) or 
        the default, which uses the absorbance that has already been
        normalized to thickness and offset to start at zero in the O-H
        stretching wavenumber range.
        
        The default baseline is between wavenumbers 3200 and 3700 cm-1.
        Change this range by changing the keywords wn_low and wn_high.
        
        The default shape of the baseline is a line (linetype='line'),
        but you can set linetype = 'quadratic', 'polynomial', or 'spline'.  
        
        For quadratics, extent of curvature is determined primarily
        by the keyword curvature, which sets how much to deviate from
        the line between wn_low and wn_high. Alternatively, set 
        force_through_wn, and a quadratic will be fit through the absorbance 
        at the wavenumber(s) input in cm-1.
        You can pass in a single wavenumber or a list of wavenumbers.
        
        For higher order wavenumbers, specify the polynomial_order, and you 
        can still use include the force_through_wn.

        linetype='spline' fits using the data on either side of the peaks: 
            one arm between spline_wn_high (default=4000) and wn_high 
            (default=3700), and the second arm between spline_wn_low
            (default=3000) and wn_low (default=3200)
        spline_kind options are 'linear', 'nearest', 'zero',
        'slinear', 'quadratic', and 'cubic' (default). 
        See documentation for scipy.interpolate.interp1d for more information.

        For noisy data, try setting abs_smear_high and low to fit to 
        average absorbances around wn_low and wn_high. 10 is usually a 
        good number.
                
        Returns baseline absorption curve. 
        """
        # get raw or normalized absorbance
        if raw_data is True:
            absorbance = self.abs_raw
        else:
            try:
                absorbance = self.abs_full_cm
            except AttributeError:
                try:
                    self.start_at_zero()
                    absorbance = self.abs_full_cm
                except AttributeError:
                    print('Spectrum has no thickness_microns attribute')
                    print('Either set raw_data=True or provide thickness')
                    return
        
        # get wavenumber range
        index_lo = (np.abs(self.wn_full-wn_low)).argmin()
        index_hi = (np.abs(self.wn_full-wn_high)).argmin()        
        base_wn = self.wn_full[index_lo:index_hi]

        # Smear start and stop over a range of wavenumbers
        abs_smear_high=int(abs_smear_high)
        abs_smear_low=int(abs_smear_low)
        if abs_high is None:
            if abs_smear_high > 0:
                gulp_hi_x = list(range(index_hi-abs_smear_high, 
                                  index_hi+abs_smear_high))
                gulp_hi_y = absorbance[gulp_hi_x]
                yhigh = np.mean(gulp_hi_y)
            else:
                yhigh = absorbance[index_hi]
        else:
            yhigh = abs_high            
        if abs_low is None:
            if abs_smear_low > 0:
                gulp_lo_x = list(range(index_lo-abs_smear_low, 
                                       index_lo+abs_smear_low))
                gulp_lo_y = absorbance[gulp_lo_x]
                ylow = np.mean(gulp_lo_y)
            else:
                ylow = absorbance[index_lo]
        else: 
            ylow = abs_low

        # start with a line 
        x = np.array([self.wn_full[index_hi], self.wn_full[index_lo]])
        y = np.array([yhigh, ylow])
        try:
            p = np.polyfit(x, y, 1)
        except ValueError:
            print(self.fname)
            print('failed to create a line across wavenumber range')
            print('wavenumbers:', x)
            print('absorbances:', y)
            return
        
        if curvature is not None or force_through_wn is not None:
            linetype = 'polynomial'
        
        if linetype == 'line':
            base_abs = np.polyval(p, base_wn)

        # make a polynomial
        elif linetype == 'polynomial':
            # add in extra points to fit through
            if force_through_wn is not None:
                if isinstance(force_through_wn, list):
                    xadd = []
                    yadd = []
                    for forcewn in force_through_wn:
                        index_mid = (np.abs(self.wn_full - forcewn)).argmin()
                        abs_at_wn_mid = absorbance[index_mid]
                        xadd.append(forcewn)
                        yadd.append(abs_at_wn_mid)
                else:                        
                    try:
                        forcewn = force_through_wn
                        index_mid = (np.abs(self.wn_full - forcewn)).argmin()
                        abs_at_wn_mid = absorbance[index_mid]
                        xadd = force_through_wn
                        yadd = abs_at_wn_mid                    
                    except TypeError:
                        print('force_through_wn must be a number, ')
                        print('the wavenumber in cm-1 you want to fit through')
                        print('or a list of wavenumbers')
                        return
            elif curvature is not None:
                yshift = curvature
                xadd = wn_mid
                yadd = np.polyval(p, wn_mid) - yshift
            x = np.insert(x, 1, xadd)
            y = np.insert(y, 1, yadd)
            
            if polynomial_order is None:
                polynomial_order = 2
            p2 = np.polyfit(x, y, polynomial_order)
            base_abs = np.polyval(p2, base_wn)

            if show_fit_values == True:
                print('fitting x values:', x)
                print('fitting y values:', y)

        # make a spline
        elif linetype == 'spline':
            idx_max = (np.abs(self.wn_full - spline_wn_high)).argmin()
            idx_min = (np.abs(self.wn_full - spline_wn_low)).argmin()
            xinterp = np.concatenate((self.wn_full[idx_min:index_lo], 
                                      self.wn_full[index_hi:idx_max]))
            yinterp = np.concatenate((absorbance[idx_min:index_lo], 
                                      absorbance[index_hi:idx_max]))
            f = interp.interp1d(xinterp, yinterp, kind=spline_kind)
            base_wn = self.wn_full[index_lo:index_hi]
            base_abs = f(base_wn)
                
        else:
            print("linetype must be 'line', 'polynomial', or 'spline'")
            return
        
        if store_baseline is True:
            self.base_mid_wn = wn_mid
            self.base_high_wn = wn_high
            self.base_low_wn = wn_low
            self.base_abs = base_abs
            self.base_wn = base_wn
        
        if show_plot is True:
            fig, ax = self.plot_spectrum(plot_raw=raw_data)
            ax.set_ylim(min(absorbance[index_lo:index_hi]), 
                        max(absorbance[index_lo:index_hi]))
            ax.set_xlim(wn_high, wn_low)
            ax.plot(base_wn, base_abs, '-k')

            if linetype == 'spline':
                ax.plot(xinterp, yinterp, 'o')
            
            if show_fit_values is True:
                if abs_smear_high > 0:
                    ax.plot(self.wn_full[gulp_hi_x], 
                            absorbance[gulp_hi_x], '-y', alpha=0.9)
                    
                if abs_smear_low > 0:
                    ax.plot(self.wn_full[gulp_lo_x], 
                            absorbance[gulp_lo_x], '-y', alpha=0.9)
                    
                if linetype == 'spline':
                    ax.plot(self.wn_full[0:index_lo], 
                            absorbance[0:index_lo], '-r')
                    ax.plot(self.wn_full[index_hi:idx_max], 
                            absorbance[index_hi:idx_max], '-r')
                else:
                    ax.plot(x, y, 'ro', alpha=0.4)
                    
        return base_abs


    def subtract_baseline(self, baseline_abs=None, wn_low=None, wn_high=None,
                          show_plot=False, raw_data=False):
        """
        Returns baseline-subtracted absorbance.
        
        First it looks for a baseline that has been passed in directly 
        through base_abs over a wavenumber range from wn_low to wn_high.
        
        Then it looks for a stored baseline in Spectrum.base_abs over
        a wavenumber range stored in Spectrum.base_low_wn and base_high_wn.
        
        If it can't find a baseline, it will make the default line between
        wavenumbers 3200 and 3700 cm-1.
        
        By default tt won't show the plot unless show_plot=True.
        """
        # get baseline absorbance
        if baseline_abs is None:
            try:
                base_abs = self.base_abs
            except AttributeError:
                print('Making default linear baseline')
                print('Use Spectrum.make_baseline for other baseline options')
                base_abs = self.make_baseline(raw_data=raw_data)
        else:
            base_abs = baseline_abs

        if base_abs is None:
            print('No baseline. Try using raw data or check thickness.')
            return

        # get wavenumber range in cm-1
        if wn_low is None:
            wn_low = np.max(self.base_wn)
        
        if wn_high is None:
            wn_high = min(self.base_wn)
            
        index_lo = (np.abs(self.wn_full-wn_low)).argmin()
        index_hi = (np.abs(self.wn_full-wn_high)).argmin()

        absorbance = self.absorbance_picker()
        if index_lo < index_hi:
            humps = absorbance[index_lo:index_hi] 
        else:
            humps = absorbance[index_hi:index_lo]
        
        # lengths sometimes don't match up exactly because of rounding
        if len(humps) > len(base_abs):
            ndif = len(humps) - len(base_abs)
            humps = humps[0:-ndif]
        elif len(humps) < len(base_abs):
            ndif = len(base_abs) - len(humps)
            for n in range(ndif):
                humps = np.append(humps, humps[-1])

        abs_nobase_cm = humps - base_abs

        if show_plot is True:
            self.plot_showbaseline()

        self.abs_nobase_cm = abs_nobase_cm
        return abs_nobase_cm


    def make_area(self, show_plot=False, raw_data=False,
                             printout=True, numformat='{:.1f}', 
                             wn_high=None, wn_low=None):
        """
        Returns area under the curve in cm^2 between wavenumbers given
        by wn_high and wn_low.
        
        If wn_high and wn_low are not set, they are assumed to be equal to 
        the full range of the baseline.
        
        If raw_data is set to True, the area will be determined for the raw
        data. The default is to determine the thickness-normalized area.
        
        By default it prints out (printout=True) the fname and area in the
        format specified by numformat, default 1 number after the decimal. 
        
        To plot the baseline at the same time, set show_plot=True.        
        """
        self.subtract_baseline(raw_data=raw_data)
        
        if wn_high is None:
            wn_high = self.base_high_wn
        if wn_low is None:
            wn_low = self.base_low_wn            
        dx = wn_high - wn_low

        idx_high = (np.abs(self.base_wn-wn_high)).argmin()
        idx_low = (np.abs(self.base_wn-wn_low)).argmin()
        dy = np.mean(self.abs_nobase_cm[idx_low:idx_high])
        
        area = dx * dy
        self.area = area
        
        if printout is True:
            print(self.fname)
            print('area:', numformat.format(area), '/cm^2')
        
        if show_plot is True:
            fig, ax = self.plot_showbaseline()
            x = self.base_wn[idx_low : idx_high]
            y1 = self.base_abs[idx_low : idx_high]
            y2 = self.abs_nobase_cm[idx_low : idx_high]
            y3 = np.array(y1) + np.array(y2)
            ax.fill_between(x, y1, y3, color='g', alpha=0.3)
            return fig, ax, area
            
        return area


    def water(self, phase='cpx', calibration='Bell', numformat='{:.1f}',
              printout=True, scale_water=3):
        """
        Returns the area under the curve and an associated water estimate 
        for a single FTIR spectrum.
        
        Takes the are under the curve for the existing baseline, and
        multiplies by the absorption coefficient for the specified phase
        (e.g., 'olivine' or default 'cpx') and calibration (default 'Bell')
        and then multiplies by the scale_water (default 3). 
        
        These estimates are not really to be trusted. See, e.g., Withers 2013.
        """
        area = self.make_area(show_plot=False, printout=False)
        w = pynams.area2water(area, phase=phase, calibration=calibration)
                        
        # output for each spectrum
        if printout is True:
            print(self.fname)
            print(''.join(('water: ', numformat.format(w), ' ppm H2O\n *',
                           str(scale_water), ' = ', 
                    numformat.format(w*3.), ' ppm H2O')))
        return w*scale_water


    def save_spectrum(self, delim='\t', file_ending='-per-cm.txt', 
                      folder=None, raw_data=False, printout=True):
        """
        Save entire spectrum divided by thickness to file with headings.
        1st column: wavenumber /cm; 2nd column: absorbance /cm.
        Default is the format ready for upload to PULI database 
        (http://puli.mfgi.hu/), but to save, e.g.,
        a spectrum that is the average of a profile, change the delimiter
        (delim), file_ending to '.CSV', and set raw_data=True if you don't
        want it to normalize to thickness if possible.
        """
        if self.fname is None:
            print('Need .fname to know what to call saved file')
            return

        if raw_data is True:
            absorbance = self.abs_raw
        else:
            try:            
                absorbance = self.abs_full_cm
            except AttributeError:
                self.divide_by_thickness()
                try:
                    absorbance = self.abs_full_cm
                except AttributeError:
                    print('Saving raw absorbances.')
                    absorbance = self.abs_raw
                    
        if folder is None:
            folder = self.folder
            
        abs_filename = ''.join((folder, self.fname, file_ending))
        a = np.transpose(np.vstack((self.wn_full, absorbance)))
        d = pd.DataFrame(data=a)
        d.to_csv(abs_filename, index=False, header=False)
        if printout is True:
            print('Saved', abs_filename)
            
            
    def save_baseline(self, folder=None, delim=',',
                      baseline_ending='-baseline.CSV'):
        """
        Save baseline (and baseline-subtracted) absorbance spectrum 
        (-baseline) to file fname-baseline.CSV in the specified folder. 
        
        The baselines can be retrieved using get_baseline(). 
        Use save_spectrum() to save full wavenumber and -per-cm absorbances,
        which in most cases will already exist but may not in the case of
        averaged or synthetic spectra.
        
        You can save multiple baselines by changing the keyword 
        baseline_ending. 
        
        The default is a CSV file, but the delimiter (delim) can be changed
        along with the file ending in baseline_ending.
        """
        if folder is None:
            folder = self.folder
            
        # getting baseline absorbance
        if self.base_abs is None:
            print('making default linear baseline')
            base_abs = self.make_baseline()
        else:
            base_abs = self.base_abs
            
        # the baseline-subtracted absorbance is also saved in this file
        abs_nobase_cm = self.subtract_baseline()
        
        # name the baseline file
        base_filename = folder + self.fname + baseline_ending
        
        # make headings in the baseline file
        t = ['wavenumber (/cm)', 'baseline value (/cm)', 
             'baseline-subtracted absorbance (/cm)']
        with open(base_filename, 'w') as base_file:
            for item in t:
                base_file.write(item+delim)
            base_file.write('\n')
            
        # write data to baseline file 
        a = np.transpose(np.vstack((self.base_wn, base_abs, abs_nobase_cm)))
        d = pd.DataFrame(data=a, columns=['baseline wavenumber (cm-1)', 
                                      'baseline absorbance (cm-1)', 
                                      'baseline-subtracted absorbance (cm-1)'])
        d.to_csv(base_filename, index=False)
        print('Saved', base_filename)


    def get_baseline(self, folder=None, delim=',', 
                     baseline_ending='-baseline.CSV',
                     print_confirmation=True):
        """
        Get baseline saved using save_baseline().
        
        Requires folder for path location and baseline_ending if other
        than '-baseline.CSV'.
        
        Returns baseline absorbances and baseline-subtracted absorbances. 
        Sets the spectrum's base_wn attribute.
        
        Set print_confirmation=False to suppress printout when file
        successfully retrieved
        """
        if folder is None:
            folder = self.folder
        try:
            filename = ''.join((folder, self.fname, baseline_ending))
        except TypeError:
            print('Problem making filename.')
            print('Probably you need to specify the folder.')
        if os.path.isfile(filename) is False:
            return
        data = pd.read_csv(filename)

        if print_confirmation is True:
            print('Got baseline ', filename)
        columns = data.columns
        self.base_wn = data[columns[0]]
        self.base_abs = data[columns[1]]
        self.abs_nobase_cm = data[columns[2]]
        self.base_high_wn = np.max(data[columns[0]])
        self.base_low_wn = np.min(data[columns[0]])
        return self.base_abs, self.abs_nobase_cm
        
    
    def get_3baselines(self, folder=None, delim=',', 
                baseline_ending='-3baselines.CSV'):
        """Returns block of baseline data saved by water_from_spectra()
        d[:,0] = baseline wavenumbers, d[:,1:4] = baselines, 
        d[:,4:7] = baseline-subtracted absorbances"""
        filename = folder + self.fname + baseline_ending
        if os.path.isfile(filename) is False:
            print(' ')
            print(self.fname)            
            print('Run water_from_spectra() with savebaselines=True')
            return
        data = np.genfromtxt(filename, delimiter=',', dtype='float', 
                             skip_header=1)
        return data                            
                
    
    def make_peakfit(self, sensitivity=40, peak_positions=None,
                     peak_heights=None, peak_widths=None,
                     show_plot=True):
        """
        Fiddle with the peakfits by passing in peak_positions in as a list of
        wavenmbers in cm-1 and/or peak_heights as a list of baseline-subtracted
        peak heights in cm-1, and/or peak_widths as a list of Gaussian peak 
        widths in wavenumbers cm-1.
        
        If there is no peak information passed in, it will guess the
        peak positions using spectrum.find_peaks(), assume a width of 50 
        for all peaks, and take the height as the absorbance at each peak
        position.
        
        If show_plot is True (default), it runs spectrum.plot_showpeakfit()
        and returns the figure and axes handles
        """
        self.subtract_baseline()
        if self.abs_nobase_cm is None:
            return

        if peak_positions is None:
            self.peakpos = self.find_peaks(show_plot=False, printout=False,
                                           sensitivity=sensitivity)
            print('peak positions:', self.peakpos)
        else:
            self.peakpos = [float(i) for i in peak_positions]
            
        
        if peak_widths is None:
            self.peak_widths = np.ones_like(self.peakpos) * 50.
        else:
            self.peak_widths = [float(i) for i in peak_widths]

        if peak_heights is None:                                           
            self.peak_heights = np.ones_like(self.peakpos)
            for idx, wn in enumerate(self.peakpos):
                idx_peak = (np.abs(self.base_wn-wn)).argmin()
                self.peak_heights[idx] = self.abs_nobase_cm[idx_peak]
            print('heights:', self.peak_heights)
        else:
            self.peak_heights = [float(i) for i in peak_heights]
        
        self.numPeaks = len(self.peakpos)
        self.make_peakareas()
        
        if show_plot is True:
            fig, ax = self.plot_peakfit()
            return fig, ax


    def save_peakfit(self, folder=None, peak_ending='-peakfit.CSV',
                     delim=','):
        """
        Save peakfit information in a file called fname+peak_ending.
        The default peak_ending is '-peakfit.CSV'. Default delimiter (delim)
        for separating values in the file is a comma.
        """
        if folder is None:
            folder = self.folder          
        filename = folder + self.fname + peak_ending
        a = np.transpose(np.vstack((self.peakpos, self.peak_heights, 
                                    self.peak_widths, self.peak_areas)))
        t = ['peak position (wavenumber /cm)', 'height (/cm)', 
             'width of gaussian (/cm)', 'area (/cm2)']
        d = pd.DataFrame(data=a, columns=t)
        d.to_csv(filename, index=False)
        print('Saved', filename)


    def make_peakfit_like(self, spectrum):
        """
        Takes another spectrum and sets this spectrum's peakfit information
        equal to that of that of the other spectrum. 
        """
        self.subtract_baseline()
        self.peakpos = spectrum.peakpos
        self.peak_heights = spectrum.peak_heights
        self.peak_widths = spectrum.peak_widths
        self.numPeaks = spectrum.numPeaks
        self.make_peakareas()


    def get_peakfit(self, folder=None, delim=',', 
                    peak_ending='-peakfit.CSV'):
        """
        Get individual peaks for spectrum from peakfit file, 
        typically fname-peakfit.CSV, but the peak_ending 
        (default='-peakfit.CSV') can be changed. 
        
        Data is stored in attributes peak_heights, peak_widths, and
        peak_areas
        """
        if folder is None:
            folder = self.folder
        filename = folder + self.fname + peak_ending
        if os.path.isfile(filename) is True:
            previous_fit = pd.read_csv(filename)
            
            # older version peakfit files don't have headings
            if len(list(previous_fit)[0]) < 20:
                previous_fit = pd.read_csv(filename, header=None)
            
            headers = list(previous_fit)
            previous_fit = previous_fit.sort_values(by=[headers[0]])
            self.peakpos = np.array(previous_fit[headers[0]])
            self.numPeaks = len(self.peakpos)
            self.peak_heights = np.array(previous_fit[headers[1]])
            self.peak_widths = np.array(previous_fit[headers[2]])
            self.peak_areas = np.array(previous_fit[headers[3]])
            print('Got peak info from', filename)
        else:
            print(' ')            
            print('Unable to get peak info from', filename)
            print('Try spec.make_peakfit')            
        
        
    def get_gaussians(self):
        """
        Generates Gaussian curves from peakfitting info. 
        
        Requires both peak fitting information and a baseline.
        
        Returns (1) peakfitcurves, all of the individual gaussians,
        and (2) summed_spectrum, the sum of all the peakfitcurvesb
        """
        try:
            peakpos = self.peakpos
        except AttributeError:
            print('No peakfitting information available yet')
            return False, False
        
        try:
            peakfitcurves = np.ones([len(peakpos), len(self.base_wn)])
        except AttributeError:
            print('Make or get a baseline before getting peakfit')
            return False, False
            
        summed_spectrum = np.zeros_like(self.base_wn)
        for k in range(len(self.peakpos)):
            peakfitcurves[k] = pynams.make_gaussian(x=self.base_wn, 
                                                pos=self.peakpos[k],
                                                h=self.peak_heights[k], 
                                                w=self.peak_widths[k])

            summed_spectrum += peakfitcurves[k]
        return peakfitcurves, summed_spectrum


    def make_peakareas(self):
        """
        Determine peak area based on Gaussian curves from peak
        width and height and store them in spectrum.peak_areas
        """
        peakfitcurves, summed_spectrum = self.get_gaussians()
        if peakfitcurves is False:
            print('Problem making Gaussian for peak area')
            return
        else:
            dx = self.base_high_wn - self.base_low_wn
            peak_areas = []
            for curve in peakfitcurves:
                peak_areas.append(dx * np.mean(curve))
            self.peak_areas = peak_areas
#            print('# of curves', len(peakfitcurves))
#            print('# of peaks', len(self.peakpos))
            
    
    def make_composite_peak(self, peak_idx_list):
        """
        Make a new 'peak', e.g., for [Ti] in olivine, by summing up 
        other peaks given by their indexes in peak_idx_list
        
        Input is a list of peak indexes to be grouped together.
        
        Adds a new peak to the list of existing information, and lists 
        the new peak position as the sum of the wavenumbers of the 
        peaks passed in to create the new composite peak information.
        """
        peakpos_new = 0.
        peak_height_new = 0.
        peak_width_new = 0.
        peak_area_new = 0.
        
        # determine new peak index and if it is already present
        for idx in peak_idx_list:
            peakpos_new = peakpos_new + self.peakpos[idx]    

        if np.in1d(peakpos_new, self.peakpos)[0]:
            print('That composite peak already exists')
            
        else:
            for idx in peak_idx_list:
                peak_height_new = peak_height_new + self.peak_heights[idx]
                peak_width_new = peak_width_new + self.peak_widths[idx]
                peak_area_new = peak_area_new + self.peak_areas[idx]
            self.peakpos = np.append(self.peakpos, peakpos_new)
            self.peak_heights = np.append(self.peak_heights, peak_height_new)
            self.peak_widths = np.append(self.peak_widths, peak_width_new)
            self.peak_areas = np.append(self.peak_areas, peak_area_new)
        
        
    def plot_peakfit(self, style=styles.style_spectrum, 
                     stylesum=styles.style_summed, 
                     axes = None, legend=True,
                     stylepeaks=styles.style_fitpeak, ytop=None, legloc=1):
        """Single spectrum: Plot peaks fit in MATLAB using peakfit.m
        REQUIRES the peak_ending and baseline_ending so that it can
        locate the file spectrum.fname+peak_ending and +baseline_ending.
        If those aren't present or the file doesn't exist, it will screw 
        up."""
        # Take baseline-subtracted spectrum from saved file every time 
        # to avoid any possible funny business from playing with baselines

        gaussian, summed_spectrum = self.get_gaussians() 
        
        if gaussian is False:
            return False, False
        
        if axes is None:
            try:
                fig, ax = self.plot_subtractbaseline(style=style, label='') 
            except AttributeError:
                print('Remember to make or get baseline')
        else: 
            ax = axes
            
        ax.plot(self.base_wn, self.abs_nobase_cm, label='observed', **style)
        ax.plot(self.base_wn, gaussian[0], label='fit bands', **stylepeaks)
        ax.plot(self.base_wn, summed_spectrum, label='peak sum', **stylesum)
        
        if axes is None:
            if legend is True:
                ax.legend(loc=legloc)
            ax.set_ylim(0., ytop)        
        
        if ytop is None:
            topnat = ax.get_ylim()[1]
            ytop = topnat + topnat*0.75
            

        for k in range(len(self.peakpos)):
            ax.plot(self.base_wn, gaussian[k], **stylepeaks)

        if axes is None:
            fig.set_size_inches(6, 6)
            ax.set_title(self.fname)
            return fig, ax


    def abs_at_given_wn(self, wn, absorbance='thickness normalized'):
        """Input wavenumber, output absorbance that is thickness normalized
        (absorbance='normalized' default), 'raw', or 'baseline-subtracted'"""

        # check you know what the wavenumbers are
        if self.wn_full is None:
            check = self.get_data()
            if check is False:
                return False

        idx = (np.abs(self.wn_full-wn)).argmin()
                
        if absorbance == 'thickness normalized':
            if self.abs_full_cm is None:
                self.start_at_zero()
            abs_at_wn = self.abs_full_cm[idx]
            return abs_at_wn
            
        elif absorbance == 'raw':
            print('Sorry, raw not programmed in yet')
            return
        elif absorbance == 'baseline-subtracted':
            print('Sorry, baseline-subtracted not programmed in yet')
            return
        else: 
            print('absorbance options are:')
            print('thickness normalized (default)')
            print('raw')
            print('baseline-subtracted')


    def absorbance_picker(self):
        """Is this raw or thickness normalized absorbance you're after?"""
        try:
            if self.thickness_microns is None:
                absorbance = self.abs_raw
            else:
                try:
                    absorbance = self.abs_full_cm
                except AttributeError:
                    self.start_at_zero()
                    absorbance = self.abs_full_cm
        except AttributeError:
            absorbance = self.abs_raw
        return absorbance
        
    
    def plot_showbaseline(self, axes=None, 
                          abs_baseline=None, 
                          wn_baseline=None, 
                          style=styles.style_spectrum, 
                          style_base=styles.style_baseline,
                          label=None, 
                          label_baseline=False,
                          offset=0.0, 
                          wn_xlim_left=4000, 
                          wn_xlim_right=3000.):
        """
        Plot FTIR spectrum and show baseline. 
        
        If axes is set, it will plot on that axes. Otherwise it will create
        and return new figure and axes handles. 
        
        If you have a curve that you want to pass in and use as a baseline,
        the absorbances go into abs_baseline, and the wavenumbers go in
        as wn_baseline. It will draw that baseline first.
        
        If no baseline is passed in explicitly, it will look for a baseline
        that has already been created by Spectrum.make_baseline, which 
        automatically saves the baseline in Spectrum.base_abs (absorbances)
        and Spectrum.base_wn (wavenumbers). 
        
        If no baseline yet exists, it will create the default linear baseline.
        
        You can pass in style dictionaries to change the line style of the
        curve (style) and the baseline (style_base). 
        
        If you later make a legend, the curve will come out labeled as the
        Spectrum.fname (default) or as the label set here. The baseline
        won't show up in the legend unless you set label_baseline=True.
        
        You can move the curve and baseline together up and down with offset.
        
        The default window is from 4000 to 3000 cm-1 wavenumbers and should
        be zoomed in on the curve. You may need to adjust your axes limits
        afterward, particularly when plotting multiple spectra.
        """
        # which axes to plot on
        if axes is None:
            fig, ax = styles.plot_spectrum_outline(wn_xlim_left=wn_xlim_left,
                                                   wn_xlim_right=wn_xlim_right)
        else:
            fig = None
            ax = axes
        
        # Get or make absorbance for baseline
        if abs_baseline is None:
            try:
                abs_baseline = self.base_abs
            except AttributeError:
                print('Making the default linear baseline.')
                print('Use Spectrum.make_baseline for other baselines.')
                abs_baseline = self.make_baseline()

        # get baseline wavenumber range
        if wn_baseline is None:
            if self.base_wn is not None:
                wn_baseline = self.base_wn
            else:
                print(('Need to pass in baseline wavenumber range too ' +
                       'either here as wn_baseline or as spectrum.base_wn'))
                return                

        # labels that would show up in a legend
        if label is None:
            label = self.fname
        style_to_use = style.copy()
        style_to_use.update({'label' : label})
        if label_baseline is True:
            style_base['label'] = 'baseline'

        # Grab thickness-normalized absorbances or raw if no thickness
        absorbance = self.absorbance_picker()

        ax.plot(self.wn_full, absorbance + offset, **style_to_use)
        ax.plot(wn_baseline, abs_baseline + offset, **style_base)

        try:        
            if self.thickness_microns is None:
                ax.set_ylabel('Raw absorbance')
        except AttributeError:
            ax.set_ylabel('Raw absorbance')

        # zoom in on y-axis
        ylow, yhigh = styles.ylim_picker(self, pad_top=0.2, pad_bot=0.2)
        yhigh = yhigh + 0.1*yhigh
        ax.set_ylim(ylow, yhigh)
            
        return fig, ax


    def plot_subtractbaseline(self, 
                              style=styles.style_spectrum, 
                              axes=None, 
                              label=None, 
                              offset=0.,):
        """
        Make and plot baseline-subtracted spectrum. style, axes, label, and
        offset are the same options as in Spectrum.plot_showbaseline()
        """
        abs_nobase_cm = self.subtract_baseline()
        
        if axes is None:
            fig, ax = styles.plot_spectrum_outline()
        else:
            fig = None
            ax = axes

        if label is None:
            label = self.fname
        style_to_use = style.copy()
        style_to_use.update({'label' : label})
            
        pad = max(abs_nobase_cm)
        yhigh = pad + 0.1*pad
        ax.set_ylim(0, yhigh)
        ax.set_xlim(self.base_high_wn, self.base_low_wn)
        ax.plot(self.base_wn, abs_nobase_cm+offset, **style_to_use)
        ax.set_title(self.fname)
        return fig, ax


    def plot_peakfit_and_baseline(self, style=styles.style_spectrum, 
                                  stylesum=styles.style_summed, 
                                  stylepeaks=styles.style_fitpeak, 
                                  style_base=styles.style_baseline,
                                  ytop=None, bottom=0., legloc=1, 
                                  label_spectrum='observed', 
                                  peak_ending='-peakfit.CSV',
                                  baseline_ending='-baseline.CSV'):
        """Plot spectrum with baseline and peakfit information together"""
        # Take baseline-subtracted spectrum from saved file every time 
        # to avoid any possible funny business from playing with baselines
        if self.peakpos is None:
            print('need to run get_peakfit first')
            return
        if self.base_wn is None:
            print('need to run get_baseline() first')
            return
        
        gaussian, summed_spectrum = self.get_gaussians(peak_ending=peak_ending, 
                                              baseline_ending=baseline_ending)
            
        wn = self.base_wn

        stylepeaks['zorder'] = 1 # otherwise peaks show up above baseline
        
        fig, ax = self.plot_spectrum(style=style, label=label_spectrum)
        ax.plot(self.base_wn, self.base_abs, label='baseline', **style_base) 
        ax.plot(wn, gaussian[0]+self.base_abs, label='fit bands', **stylepeaks)
        ax.plot(self.base_wn, summed_spectrum+self.base_abs, label='peak sum', 
                **stylesum)
                
        ax.legend(loc=legloc)
        
        if ytop is None:
            topnat = ax.get_ylim()[1]
            ytop = topnat + topnat*0.75
            
        ax.set_ylim(bottom, ytop)

        for k in range(len(self.peakpos)):
            ax.plot(self.base_wn, gaussian[k]+self.base_abs, **stylepeaks)

        return fig, ax


def make_filenames(folder, classname=Spectrum, file_ending='.CSV'):
    """ Set filename attribute based on folder and fname attribute
    for all Spectra() with fname but no filename."""
    for obj in gc.get_objects():
        if isinstance(obj, classname):
            try:
                if obj.fname is not None and obj.filename is None:
                    obj.filename = ''.join((folder, obj.fname, file_ending))
            except AttributeError:
                print('just chill, ok?')


def water_from_spectra(list3, folder, phase='cpx', 
                       proper3=False, numformat='{:.0f}',
                       savebaselines=False, show_plots=True, 
                       baseline_ending='-3baselines.CSV', delim=',', 
                       calibration='Bell', main_yshift=None,
                       window_large=None, window_small=None, yhigh=1.):
    """Produce water estimate from list of FTIR spectra; 
    Default calibration is Bell et al. 1995, ideally using 
    3 spectra that are polarized in orthogonal directions (proper3=True)
    Default (proper3=False) is to estimate area and total water from each spectrum.
    Then average and std the results."""
    if calibration == 'Bell':
        print('Bell calibration')
        pass
    elif calibration == 'Paterson':
        print('Paterson calibration not available quite yet...')
        return
    else:
        print('Only Bell (default) or Paterson calibration so far')
        return
        
    if proper3 == True and len(list3) != 3:
        print(' ')
        print('For proper3=True, list should contain only 3 spectra')
        proper3 = False
    
    uarea_list = []
    uwater_list = []
    
    for spec in list3:
        if show_plots is True:
            print('Showing spectrum for', spec.fname)
            fig, ax = spec.plot_spectrum_outline()
            plt.plot(spec.wn_full, spec.abs_full_cm, **styles.style_spectrum) 

        # how much to shift each quadratic line
        if main_yshift is None:        
            main_yshift = spec.base_mid_yshift        
        if window_large is None:
            window_large = spec.base_w_large
        if window_small is None:
            window_small = spec.base_w_small

        if spec.abs_full_cm is None:
            spec.start_at_zero()
        # Generate list of 3 possible areas under the curve
        area_list = np.array([])
        spec.make_baseline()
        baseline_list = np.ones([3, len(spec.base_wn)])
        pek_list = np.ones([3, len(spec.base_wn)])
        
        k = 0
        for bam in [window_small, -window_small, -window_large]:
            # Setting up 3 different baseline shapes
            main_yshift = main_yshift - bam
            base_abs = spec.make_baseline('quadratic', shiftline=main_yshift)
            abs_nobase_cm = spec.subtract_baseline(bline=base_abs)
            if show_plots is True:
                ax.plot(spec.base_wn, base_abs, **styles.style_baseline)

            area = spec.make_area(area_plot=False, printout=False)
            area_list = np.append(area_list, area)
            baseline_list[k,:] = base_abs
            pek_list[k,:] = abs_nobase_cm
            k+=1

        # list areas and water with uncertainties
        uarea = ufloat(np.mean(area_list), np.std(area_list))
        uarea_list.append(uarea)
        uwater = pynams.area2water(uarea, phase=phase)

        # Water estimate depends on if you are adding 3 areas 
        # or just multiplying by three and later averaging
        if proper3 is False:
            uwater = 3*uwater
        uwater_list.append(uwater)

        # I show() the figures explicitly in this first loop because
        # otherwise they show up inline *after* all the printed output
        if show_plots is True:
            if min(base_abs) > 0:
                ylow = 0
            else:
                ylow = min(base_abs) + 0.1*min(base_abs)
            plt.ylim(ylow, yhigh)
            plt.show(fig)

        if savebaselines is True:
            base_filename = folder + spec.fname + baseline_ending
            t = ['wavenumber (/cm)', 'upper baseline (/cm) (smaller area)', 
                 'main baseline (/cm)', 'lower baseline (/cm) (larger area)',
                 'absorbance - upper baseline (/cm)', 
                 'absorbance - main baseline (/cm)',
                 'absorbance - lower baseline (/cm)']
            with open(base_filename, 'w') as base_file:
                for item in t:
                    base_file.write(item+delim)
                base_file.write('\n')
            a = np.transpose(np.vstack((spec.base_wn, baseline_list, pek_list)))
            with open(base_filename, 'a') as base_file:
                np.savetxt(base_file, a, delimiter=delim)
            
    # output for each spectrum
    for x in range(len(list3)):
        print(list3[x].fname)
        print("area:", numformat.format(uarea_list[x]), "/cm^2")
        print("water:", numformat.format(uwater_list[x]), "ppm H2O")
        if savebaselines is True:
            print('Saved baselines to', list3[x].fname+baseline_ending)
        print(' ')
    # Final output
    if proper3 == True:
        a = np.sum(uarea_list)
        w = np.sum(uwater_list)
        print('Sum of individual contributions')
    else:
        a = np.mean(uarea_list)
        w = np.mean(uwater_list)
        print('Averaged individual (*3) estimates')
    print('area:', numformat.format(a), '/cm^2')
    print('water:', numformat.format(w), 'ppm H2O')
    print(' ')
    return a, w  


def list_with_attribute(classname, attributename, attributevalue):
    """Gather all instances in specified class with a particular attribute"""
    my_list = []
    for obj in gc.get_objects():
        if isinstance(obj, classname):
            if getattr(obj, attributename) == attributevalue:
                my_list.append(obj.fname)
    return my_list




# -*- coding: utf-8 -*-
"""
Created on Wed Nov 04 17:50:54 2015

@author: Ferriss

pynams setup file
"""
from setuptools import setup


setup(
    description = "python for water in nominally anhydrous minerals",
    author = 'Elizabeth Ferriss',
    author_email = 'elizabeth.ferriss@gmail.com',
    url = 'https://github.com/EFerriss/Pynams',
    version = "0.2.2",
    packages = ['pynams', 'pynams.diffusion'],
    name = "pynams",
    license = "MIT",
    install_requires = ['lmfit==0.8.0', 'uncertainties>=3.1.3'],
    include_package_data = True
    )